import os
import random
import socket
import string
import gc, inspect
import ipynbname
import typing

sentry_inited = False


def init_sentry():
    global sentry_inited
    if sentry_inited:
        return
    sentry_inited = True

    try:
        import sentry_sdk
        from sentry_sdk.integrations.logging import LoggingIntegration
    except ImportError:
        return
    # Disable logs going to Sentry. Its slow
    sentry_sdk.init(integrations=[LoggingIntegration(level=None, event_level=None)])


def raise_exception_with_sentry_if_available(
    err: Exception, fingerprint: typing.Any
) -> typing.NoReturn:
    init_sentry()
    try:
        import sentry_sdk
    except ImportError:
        raise err
    else:
        with sentry_sdk.push_scope() as scope:
            scope.fingerprint = fingerprint
            # I (Tim) don't think we need to explicitly capture the exception
            # here, since we're raising it anyway. Explicitly capturing it
            # ends dropping the stack trace in Sentry.
            # sentry_sdk.capture_exception(err)
            raise err


def capture_exception_with_sentry_if_available(
    err: Exception, fingerprint: typing.Any
) -> typing.Union[None, str]:
    init_sentry()
    try:
        import sentry_sdk
    except ImportError:
        pass
    else:
        with sentry_sdk.push_scope() as scope:
            scope.fingerprint = fingerprint
            return sentry_sdk.capture_exception(err)
    return None


def get_notebook_name():
    return ipynbname.name()


def get_hostname():
    return socket.gethostname()


def get_pid():
    return os.getpid()


def rand_string_n(n: int) -> str:
    return "".join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(n)
    )


def parse_boolean_env_var(name: str) -> bool:
    return os.getenv(name, "False").lower() in ("true", "1", "t")


def parse_number_env_var(name: str) -> typing.Optional[typing.Union[int, float]]:
    raw_val = os.getenv(name)
    if raw_val is None:
        return None
    try:
        return int(raw_val)
    except ValueError:
        return float(raw_val)


def find_names(obj):
    if hasattr(obj, "name"):
        return [obj.name]
    frame = inspect.currentframe()
    for frame in iter(lambda: frame.f_back, None):
        frame.f_locals
    obj_names = []
    for referrer in gc.get_referrers(obj):
        if isinstance(referrer, dict):
            for k, v in referrer.items():
                if v is obj:
                    obj_names.append(k)
    return obj_names


def is_colab():
    try:
        import google.colab
    except ImportError:
        return False
    return True


def is_notebook():
    if is_colab():
        return True
    try:
        from IPython import get_ipython
    except ImportError:
        return False
    else:
        ip = get_ipython()
        if ip is None:
            return False
        if "IPKernelApp" not in ip.config:
            return False
        # if "VSCODE_PID" in os.environ:
        #     return False
    return True


def is_pandas_dataframe(obj):
    try:
        import pandas as pd
    except ImportError:
        return False
    return isinstance(obj, pd.DataFrame)


def _resolve_path(path: str, current_working_directory: str) -> list[str]:
    if not os.path.isabs(path):
        path = os.path.join(current_working_directory, path)
    path_parts: list[str] = []
    for part in path.split(os.path.sep):
        if part == "..":
            if path_parts:
                path_parts.pop()
        elif part != "." and part:
            path_parts.append(part)
    return path_parts


def relpath_no_syscalls(
    target_path: str, start_path: str, current_working_directory: str
) -> str:
    target_parts = _resolve_path(target_path, current_working_directory)
    start_parts = _resolve_path(start_path, current_working_directory)

    if target_parts == start_parts:
        return "."

    common_length = 0
    for target_part, start_part in zip(target_parts, start_parts):
        if target_part != start_part:
            break
        common_length += 1

    relative_parts = [".."] * (len(start_parts) - common_length) + target_parts[
        common_length:
    ]
    return os.path.sep.join(relative_parts)


def sample_rows(data: list, max_rows: int) -> list:
    data_len = len(data)

    if data_len <= max_rows:
        return data

    if max_rows <= 0:
        return []
    if max_rows == 1:
        return [data[0]]
    if max_rows == 2:
        return [data[0], data[-1]]

    split_size = max_rows // 3
    gap_size = (data_len - max_rows) // 2
    start_split = data[:split_size]
    middle_start = split_size + gap_size
    middle_split = data[middle_start : middle_start + split_size]
    end_split = data[-split_size:]
    return start_split + middle_split + end_split



from . import wb_domain_types as wdt
from .wandb_domain_gql import (
    gql_prop_op,
    gql_connection_op,
    gql_root_op,
)
from .. import weave_types as types

# Section 1/6: Tag Getters
#

# Section 2/6: Root Ops
org = gql_root_op(
    "root-org",
    "organization",
    wdt.OrgType,
    {
        "orgName": types.String(),
    },
    lambda inputs: f'name: {inputs["orgName"]}',
)

# Section 3/6: Attribute Getters
gql_prop_op("org-name", wdt.OrgType, "name", types.String())


# Section 4/6: Direct Relationship Ops


# Section 5/6: Connection Ops
gql_connection_op(
    "org-projects",
    wdt.OrgType,
    "projects",
    wdt.ProjectType,
    {},
    lambda inputs: "first: 100",
)

gql_connection_op(
    "org-reports",
    wdt.OrgType,
    "views",
    wdt.ReportType,
    {},
    lambda inputs: "first: 100",
)

gql_connection_op(
    "org-artifacts",
    wdt.OrgType,
    "artifactCollections",
    wdt.ArtifactCollectionType,
    {},
    lambda inputs: "first: 100",
)

# Section 6/6: Non Standard Business Logic Ops
#



import dataclasses
import typing

import weave
from .. import panel
from .. import panel_util
from .. import graph

CardContentType = typing.TypeVar("CardContentType")


@weave.type()
class CardTab(typing.Generic[CardContentType]):
    name: str
    content: CardContentType


@weave.type()
class CardConfig:
    title: weave.Node[str]
    subtitle: str
    content: list[CardTab]


@weave.type()
class Card(panel.Panel):
    id = "Card"
    config: typing.Optional[CardConfig] = dataclasses.field(
        default_factory=lambda: None
    )

    def __init__(self, input_node=graph.VoidNode(), vars=None, config=None, **options):
        super().__init__(input_node=input_node, vars=vars)
        self.config = config
        if self.config is None:
            self.config = CardConfig(
                panel_util.make_node(options["title"]),
                options["subtitle"],
                options["content"],
            )



import re
import random
import typing

import logging
import contextvars
import contextlib


from . import value_or_error
from . import debug_compile


from . import serialize
from . import compile_domain
from . import op_args
from . import weave_types as types
from . import graph
from . import registry_mem
from . import dispatch
from . import graph_debug
from . import stitch
from . import compile_table
from . import weave_internal
from . import engine_trace
from . import errors


# These call_* functions must match the actual op implementations.
# But we don't want to import the op definitions themselves here, since
# those depend on the decorators, which aren't defined in the engine.

DEBUG_COMPILE = False


def _dispatch_error_is_client_error(
    op_name: str, input_types: dict[str, types.Type]
) -> bool:
    from .ops_domain import wbmedia

    if op_name in set(
        (
            "tag-run",
            "count",
            "file-table",
            "offset",
            "typedDict-pick",
            "list-createIndexCheckpointTag",
            "entity-name",
            "concat",
            "string-isNumeric",
            "project-artifacts",
            "panel-table",
        )
    ):
        # All the cases we've seen of this lately are clearly client errors, so
        # we'll send back a 400!
        return True
    elif op_name == "file-path" and types.optional(
        wbmedia.ImageArtifactFileRefType()
    ).assign_type(input_types["file"]):
        # You shouldn't be able to call file-path on ImageArtifactFileRef.
        return True
    return False


def _call_run_await(run_node: graph.Node) -> graph.OutputNode:
    return graph.OutputNode(
        run_node.type.output_type.output, "run-await", {"self": run_node}  # type: ignore
    )


# We don't want to import the op definitions themselves here, since
# those depend on the decorators, which aren't defined in the engine.
def _call_execute(function_node: graph.Node) -> graph.OutputNode:
    function_node_type = typing.cast(types.Function, function_node.type)
    if isinstance(function_node, graph.ConstNode) and isinstance(
        function_node.val.type, types.Function
    ):
        return _call_execute(function_node.val)
    return graph.OutputNode(
        function_node_type.output_type, "execute", {"node": function_node}
    )


def _quote_node(node: graph.Node) -> graph.Node:
    return weave_internal.const(node)


def _dispatch_map_fn_refining(node: graph.Node) -> typing.Optional[graph.OutputNode]:
    if isinstance(node, graph.OutputNode):
        from_op = node.from_op
        try:
            op = dispatch.get_op_for_inputs(node.from_op.name, from_op.input_types)
            params = from_op.inputs
            if isinstance(op.input_type, op_args.OpNamedArgs):
                params = {
                    k: n
                    for k, n in zip(op.input_type.arg_types, from_op.inputs.values())
                }
            res = op.lazy_call(**params)
            # logging.info("Dispatched (refine): %s -> %s", node, res.type)
            return res
        except errors.WeaveDispatchError:
            logging.error(
                "Error while dispatching (refine phase)\n!=!=!=!=!\nName: %s\nInput Types: %s\nExpression: %s",
                from_op.name,
                from_op.input_types,
                re.sub(r'[\\]+"', '"', graph_debug.node_expr_str_full(node)),
            )
            if _dispatch_error_is_client_error(from_op.name, from_op.input_types):
                raise errors.WeaveBadRequest(
                    "Error while dispatching: %s. This is most likely a client error"
                    % from_op.name
                )
            else:
                raise
        except:
            logging.error(
                "Error while dispatching (refine phase)\n!=!=!=!=!\nName: %s\nInput Types: %s\nExpression: %s",
                from_op.name,
                from_op.input_types,
                re.sub(r'[\\]+"', '"', graph_debug.node_expr_str_full(node)),
            )
            raise

    return None


def _remove_optional(t: types.Type) -> types.Type:
    if types.is_optional(t):
        return types.non_none(t)
    return t


def _dispatch_map_fn_no_refine(node: graph.Node) -> typing.Optional[graph.OutputNode]:
    if isinstance(node, graph.OutputNode):
        if node.from_op.name == "tag-indexCheckpoint":
            # I'm seeing that there is no indexCheckpoint tag present
            # on types that come from WeaveJS (at least by the time we call
            # this op). Maybe a WeaveJS bug?
            # TODO
            return node
        if node.from_op.name == "file-type":
            # since we didn't refine, the input to file-type is not correct yet.
            # if its in the graph, just trust that's what we want
            # TODO: does this work for mapped case?
            return node
        from_op = node.from_op
        try:
            op = dispatch.get_op_for_inputs(node.from_op.name, from_op.input_types)
        except errors.WeaveDispatchError as e:
            if _dispatch_error_is_client_error(from_op.name, from_op.input_types):
                raise errors.WeaveBadRequest(
                    "Error while dispatching (no refine phase): %s. This is most likely a client error"
                    % from_op.name
                )
            else:
                raise

        params = from_op.inputs
        if isinstance(op.input_type, op_args.OpNamedArgs):
            params = {
                k: n for k, n in zip(op.input_type.arg_types, from_op.inputs.values())
            }

        output_type = node.type
        # In the case where we are dispatching to a new op, we want to use the
        # new op's `unrefined_output_type_for_params` output type - rather than
        # blindly trusting the client type.
        if not node.from_op.name.startswith("local-artifact://") and (
            node.from_op.name != op.name
        ):
            output_type = op.unrefined_output_type_for_params(params)

        res = graph.OutputNode(_remove_optional(output_type), op.uri, params)
        # logging.info("Dispatched (no refine): %s -> %s", node, res.type)
        return res
    return None


def _simple_optimizations(node: graph.Node) -> typing.Optional[graph.Node]:
    # Put simple graph transforms here!
    if not isinstance(node, graph.OutputNode):
        return None
    if node.from_op.friendly_name == "merge":
        # Merging two dicts where one is empty returns the other. The frontend
        # sends this pattern down a lot right now, and it causes us to break out
        # of arrow vectorization.
        lhs, rhs = node.from_op.inputs.values()
        if (
            isinstance(lhs, graph.OutputNode)
            and lhs.from_op.friendly_name == "dict"
            and not lhs.from_op.inputs
        ):
            return rhs
        if (
            isinstance(rhs, graph.OutputNode)
            and rhs.from_op.friendly_name == "dict"
            and not rhs.from_op.inputs
        ):
            return lhs
    return None


def _required_const_input_names(node: graph.OutputNode) -> typing.Optional[list[str]]:
    res = compile_domain.required_const_input_names(node)
    if res is not None:
        return res
    return None


def _resolve_required_consts(node: graph.Node) -> typing.Optional[graph.Node]:
    # Put simple graph transforms here!
    if not isinstance(node, graph.OutputNode):
        return None
    required_const_input_names = _required_const_input_names(node)
    if required_const_input_names is None:
        return None
    new_inputs = dict(node.from_op.inputs)
    for input_name, input_node in node.from_op.inputs.items():
        if input_name in required_const_input_names:
            if not isinstance(input_node, graph.ConstNode):
                result = weave_internal.use(_compile([input_node])[0])
                new_inputs[input_name] = graph.ConstNode(input_node.type, result)
    return graph.OutputNode(
        node.type,
        node.from_op.name,
        new_inputs,
    )


def _make_auto_op_map_fn(when_type: typing.Callable[[types.Type], bool], call_op_fn):
    def fn(node: graph.Node) -> typing.Optional[graph.Node]:
        if isinstance(node, graph.OutputNode):
            node_inputs = node.from_op.inputs
            op_def = registry_mem.memory_registry.get_op(node.from_op.name)
            if (
                op_def.name == "tag-indexCheckpoint"
                # or op_def.name == "Object-__getattr__"
                or op_def.name == "set"
                # panel_scatter and panel_distribution have the incorrect
                # input types for their config arg. They should be weave.Node.
                # We need a frontend fix to handle that. For now there's a hack
                # here.
                # TODO: Fix in frontend and panel_* and remove this hack.
                or (
                    isinstance(op_def.concrete_output_type, types.Type)
                    and op_def.concrete_output_type._base_type is not None
                    and op_def.concrete_output_type._base_type.name == "Panel"
                )
            ):
                # These are supposed to be a passthrough op, we don't want to convert
                # it. TODO: Find a more general way, maybe by type inspection?
                return None
            new_inputs: dict[str, graph.Node] = {}
            swapped = False
            for k, input_node in node_inputs.items():
                actual_input_type = input_node.type
                new_inputs[k] = input_node
                if not when_type(actual_input_type):
                    continue
                if isinstance(op_def.input_type, op_args.OpNamedArgs):
                    op_input_type = op_def.input_type.arg_types[k]
                elif isinstance(op_def.input_type, op_args.OpVarArgs):
                    op_input_type = op_def.input_type.arg_type
                else:
                    raise ValueError(
                        f"Unexpected op input type {op_def.input_type} for op {op_def.name}"
                    )
                if callable(op_input_type):
                    continue
                if not when_type(op_input_type):
                    new_inputs[k] = call_op_fn(input_node)
                    swapped = True
            if swapped:
                return graph.OutputNode(node.type, node.from_op.name, new_inputs)
        return None

    return fn


def _make_inverse_auto_op_map_fn(when_type: type[types.Type], call_op_fn):
    def fn(node: graph.Node) -> typing.Optional[graph.Node]:
        if isinstance(node, graph.OutputNode):
            node_inputs = node.from_op.inputs
            op_def = registry_mem.memory_registry.get_op(node.from_op.name)
            new_inputs: dict[str, graph.Node] = {}
            for k, input_node in node_inputs.items():
                actual_input_type = input_node.type
                new_inputs[k] = input_node
                if isinstance(actual_input_type, when_type):
                    continue
                if isinstance(op_def.input_type, op_args.OpNamedArgs):
                    op_input_type = op_def.input_type.arg_types[k]
                elif isinstance(op_def.input_type, op_args.OpVarArgs):
                    op_input_type = op_def.input_type.arg_type
                else:
                    raise ValueError(
                        f"Unexpected op input type {op_def.input_type} for op {op_def.name}"
                    )
                if callable(op_input_type):
                    continue
                if isinstance(op_input_type, when_type):
                    new_inputs[k] = call_op_fn(input_node)
            return graph.OutputNode(node.type, node.from_op.name, new_inputs)
        return None

    return fn


_await_run_outputs_map_fn = _make_auto_op_map_fn(
    lambda t: isinstance(t, types.Function)
    and isinstance(t.output_type, types.RunType),
    _call_run_await,
)

_execute_nodes_map_fn = _make_auto_op_map_fn(
    lambda t: isinstance(types.split_none(t)[1], types.Function), _call_execute
)

_quote_nodes_map_fn = _make_inverse_auto_op_map_fn(types.Function, _quote_node)


def compile_apply_column_pushdown(
    leaf_nodes: list[graph.Node], on_error: graph.OnErrorFnType = None
) -> list[graph.Node]:
    # This is specific to project-runs2 (not yet used in W&B production) for now. But it
    # is a general pattern that will work for all arrow tables.

    op_names = [
        "project-runs2",
        "run-history2",
        "mapped_run-history2",
        "run-history",
        "mapped_run-history",
        "run-history3",
        "mapped_run-history3",
    ]

    if not graph.filter_nodes_full(
        leaf_nodes,
        lambda n: isinstance(n, graph.OutputNode) and n.from_op.name in op_names,
    ):
        return leaf_nodes

    p = stitch.stitch(leaf_nodes, on_error=on_error)

    def _replace_with_column_pushdown(node: graph.Node) -> graph.Node:
        if isinstance(node, graph.OutputNode) and node.from_op.name in op_names:
            forward_obj = p.get_result(node)
            run_cols = compile_table.get_projection(forward_obj)
            if node.from_op.name == "project-runs2":
                config_cols = list(run_cols.get("config", {}).keys())
                summary_cols = list(run_cols.get("summary", {}).keys())
                return graph.OutputNode(
                    node.type,
                    "project-runs2_with_columns",
                    {
                        "project": node.from_op.inputs["project"],
                        "config_cols": weave_internal.const(config_cols),
                        "summary_cols": weave_internal.const(summary_cols),
                    },
                )
            if "run-history" in node.from_op.name:
                history_cols = list(run_cols.keys())

                if len(history_cols) > 0:
                    return graph.OutputNode(
                        node.type,
                        node.from_op.name + "_with_columns",
                        {
                            "run": node.from_op.inputs["run"],
                            "history_cols": weave_internal.const(
                                list(set([*history_cols, "_step"]))
                            ),
                        },
                    )
        return node

    return graph.map_nodes_full(leaf_nodes, _replace_with_column_pushdown, on_error)


def compile_fix_calls(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    return graph.map_nodes_full(nodes, _dispatch_map_fn_no_refine, on_error)


def compile_simple_optimizations(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    return graph.map_nodes_full(nodes, _simple_optimizations, on_error)


def compile_resolve_required_consts(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    return graph.map_nodes_full(nodes, _resolve_required_consts, on_error)


def compile_await(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    return graph.map_nodes_full(nodes, _await_run_outputs_map_fn, on_error)


def compile_execute(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    # Actually does the execution here in compile phase.
    # I made this change to handle cases where we need to pass mutations through
    # execute calls, which happens when we have Nodes stored in Const nodes (panels)
    # that we are mutating.
    # However I think I later solved this with client-side execution and it can maybe
    # be removed.
    # I'm leaving this for now as it doesn't affect W&B prod (which never calls execute).
    with_execute_ops = graph.map_nodes_full(nodes, _execute_nodes_map_fn, on_error)
    return with_execute_ops

    def _replace_execute(node: graph.Node) -> typing.Optional[graph.Node]:
        if isinstance(node, graph.OutputNode) and node.from_op.name == "execute":
            res = weave_internal.use(node.from_op.inputs["node"])
            if not isinstance(res, graph.Node):
                raise ValueError(
                    f"Expected node to be a Node, got {res} of type {type(res)}"
                )
            return compile_fix_calls([res])[0]
        return None

    return graph.map_nodes_full(with_execute_ops, _replace_execute)


def _resolve_function_calls(node: graph.Node) -> typing.Optional[graph.Node]:
    if (
        not isinstance(node, graph.OutputNode)
        or node.from_op.name != "function-__call__"
    ):
        return node

    inputs = list(node.from_op.inputs.values())
    fn_node = inputs[0]
    if not (
        isinstance(fn_node, graph.ConstNode)
        and isinstance(fn_node.type, types.Function)
    ):
        return node

    while isinstance(fn_node.val, graph.ConstNode) and isinstance(
        fn_node.type, types.Function
    ):
        fn_node = fn_node.val

    return weave_internal.better_call_fn(fn_node, *inputs[1:])


def compile_function_calls(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    return graph.map_nodes_full(nodes, _resolve_function_calls, on_error)


def compile_quote(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    return graph.map_nodes_full(nodes, _quote_nodes_map_fn, on_error)


def compile_refine(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    return graph.map_nodes_full(nodes, _dispatch_map_fn_refining, on_error)


def compile_merge_by_node_id(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    seen_nodes: dict[str, graph.Node] = {}

    def _merge_by_node_id(node: graph.Node) -> graph.Node:
        try:
            nid = serialize.node_id(node)
        except TypeError:
            nid = str(random.random())
        if nid in seen_nodes:
            return seen_nodes[nid]
        if isinstance(node, graph.OutputNode):
            node = dispatch.RuntimeOutputNode(
                node.type, node.from_op.name, node.from_op.inputs
            )
        elif isinstance(node, graph.VarNode):
            node = dispatch.RuntimeVarNode(node.type, node.name)
        elif isinstance(node, graph.ConstNode):
            node = dispatch.RuntimeConstNode(node.type, node.val)
        seen_nodes[nid] = node
        return node

    return graph.map_nodes_full(nodes, _merge_by_node_id, on_error)


def _node_ops(node: graph.Node) -> typing.Optional[graph.Node]:
    if not isinstance(node, graph.OutputNode):
        return None
    if node.from_op.name not in [
        "RunChain-history",
        "panel_table-active_data",
        "Query-selected",
        "panel_plot-selected_data",
        "panel_table-all_rows",
    ]:
        return None
    new_node = typing.cast(graph.Node, weave_internal.use(node))

    # The result will typically contain expressions that were sent down as
    # part of panel configs within Const nodes. They haven't been handled
    # by deserialize/compile yet.
    new_node_fixed_calls = compile_fix_calls([new_node])[0]

    # This last line dedupes nodes that are identical. But it doesn't
    # globally dedupe against the graph we've already compiled.
    # There is still some duplication left in Weave execution...
    new_node_merged = compile_merge_by_node_id([new_node_fixed_calls])[0]

    # Do it again to handle nested calls
    return compile_node_ops([new_node_merged])[0]


def compile_node_ops(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    return graph.map_nodes_full(nodes, _node_ops, on_error)


# This compile pass using the `top_level` mapper since we recurse manually. We can't use
# the full mapper, because we need to traverse when encountering the lambda op itself,
# rather than the const node since we need uniqueness based on the lambda op.
def compile_lambda_uniqueness(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    return graph.map_nodes_top_level(nodes, _compile_lambda_uniqueness, on_error)


def _compile_lambda_uniqueness(node: graph.Node) -> typing.Optional[graph.Node]:
    # In the case that the node has a lambda op, then we recurse into any
    # lambda parameter. If the result is a new output var, then we update the
    # current node, ensuring that every lambda parameter is unique to the op.
    #
    # There is another benefit of this approach: parts of the lambda graph which are
    # shared between lambdas AND don't have a var ancestor will be shared in memory.
    # This means that if a lambda has a shared "non-dependent" parameter, then it will
    # only be executed once, and will be treated as the same logical op in stitch operations.
    # This is the desired behavior.
    if isinstance(node, graph.OutputNode):
        new_inputs = {}
        for input_key, input_node in node.from_op.inputs.items():
            if isinstance(input_node, graph.ConstNode) and isinstance(
                input_node.val, graph.Node
            ):
                uniq_lambda = compile_lambda_uniqueness([input_node.val])[0]
                if uniq_lambda is not input_node.val:
                    new_inputs[input_key] = graph.ConstNode(
                        input_node.type, uniq_lambda
                    )
        if len(new_inputs) > 0:
            use_inputs: dict[str, graph.Node] = {
                k: new_inputs.get(k, v) for k, v in node.from_op.inputs.items()
            }
            return weave_internal.make_output_node(
                node.type, node.from_op.name, use_inputs
            )

    # This is where the magic happens. We need to ensure that var nodes
    # are unique in memory
    elif isinstance(node, graph.VarNode):
        return weave_internal.make_var_node(node.type, node.name)

    return node


const_node = weave_internal.make_const_node(types.NoneType(), None)


def _track_errors(fn):
    def final(
        nodes: typing.List[graph.Node],
    ) -> value_or_error.ValueOrErrors[graph.Node]:
        compile_errors = []

        def on_error(ndx: int, e: Exception):
            compile_errors.append((ndx, e))
            return const_node

        inner_res = fn(nodes, on_error)

        results: list[value_or_error.ValueOrError[graph.Node]] = [
            value_or_error.Value(node) for node in inner_res
        ]
        for ndx, e in compile_errors:
            results[ndx] = value_or_error.Error(e)

        return value_or_error.ValueOrErrors(results)

    return final


def _compile(
    nodes: typing.List[graph.Node],
) -> value_or_error.ValueOrErrors[graph.Node]:
    tracer = engine_trace.tracer()
    # logging.info("Starting compilation of graph with %s leaf nodes" % len(nodes))

    results = value_or_error.ValueOrErrors.from_values(nodes)

    # If we're being called from WeaveJS, we need to use dispatch to determine
    # which ops to use. Critically, this first phase does not actually refine
    # op output types, so after this, the types in the graph are not yet correct.
    with tracer.trace("compile:fix_calls"):
        results = results.batch_map(_track_errors(compile_fix_calls))

    with tracer.trace("compile:simple_optimizations"):
        results = results.batch_map(_track_errors(compile_simple_optimizations))

    with tracer.trace("compile:lambda_uniqueness"):
        results = results.batch_map(_track_errors(compile_lambda_uniqueness))

    # Auto-transforms, where we insert operations to convert between types
    # as needed.
    # TODO: is it ok to have this before final refine?
    with tracer.trace("compile:await"):
        results = results.batch_map(_track_errors(compile_await))
    with tracer.trace("compile:execute"):
        results = results.batch_map(_track_errors(compile_execute))
    with tracer.trace("compile:function_calls"):
        results = results.batch_map(_track_errors(compile_function_calls))
    with tracer.trace("compile:quote"):
        results = results.batch_map(_track_errors(compile_quote))

    # Some ops require const input nodes. This pass executes any branches necessary
    # to ensure that requirement holds.
    # Only gql ops require this for now.
    with tracer.trace("compile:resolve_required_consts"):
        results = results.batch_map(_track_errors(compile_resolve_required_consts))

    with tracer.trace("compile:node_ops"):
        results = results.batch_map(_track_errors(compile_node_ops))

    # Now that we have the correct calls, we can do our forward-looking pushdown
    # optimizations. These do not depend on having correct types in the graph.

    with tracer.trace("compile:gql"):
        results = results.batch_map(
            _track_errors(compile_domain.apply_domain_op_gql_translation)
        )

    with tracer.trace("compile:column_pushdown"):
        results = results.batch_map(_track_errors(compile_apply_column_pushdown))

    # Final refine, to ensure the graph types are exactly what Weave python
    # produces. This phase can execute parts of the graph. It's very important
    # that this is the final phase, so that when we execute the rest of the
    # graph, we reuse any results produced in this phase, instead of re-executing
    # those nodes.
    with tracer.trace("compile:refine"):
        results = results.batch_map(_track_errors(compile_refine))

    # This is very expensive!
    # loggable_nodes = graph_debug.combine_common_nodes(n)
    # logging.info(
    #     "Compilation complete. Result nodes:\n%s",
    #     "\n".join(graph_debug.node_expr_str_full(n) for n in loggable_nodes),
    # )

    if DEBUG_COMPILE:
        debug_compile.check_weave0_compile_result(
            nodes,
            [item if err == None else const_node for item, err in results.iter_items()],
        )

    return results


_compile_disabled: contextvars.ContextVar[bool] = contextvars.ContextVar(
    "_compile_disabled", default=False
)


def _is_compiling() -> bool:
    return _compile_disabled.get()


@contextlib.contextmanager
def _compiling():
    token = _compile_disabled.set(True)
    try:
        yield
    finally:
        _compile_disabled.reset(token)


@contextlib.contextmanager
def enable_compile():
    token = _compile_disabled.set(False)
    try:
        yield
    finally:
        _compile_disabled.reset(token)


def compile(
    nodes: typing.List[graph.Node],
) -> value_or_error.ValueOrErrors[graph.Node]:
    """
    This method is used to "compile" a list of nodes. Here we can add any
    optimizations or graph rewrites
    """
    # The refine phase may execute parts of the graph. Executing recursively
    # calls compile. Use context to ensure we only compile the top level
    # graph.
    if _is_compiling():
        return value_or_error.ValueOrErrors.from_values(nodes)
    with _compiling():
        return _compile(nodes)



from weave import context_state as _context

_loading_builtins_token = _context.set_loading_built_ins()

from . import scratch
from . import eval
from . import petdataset

_context.clear_loading_built_ins(_loading_builtins_token)



import contextlib
import dataclasses
import hashlib
import os
import json
import typing
import shutil
from datetime import datetime
import pathlib
import tempfile

from . import uris
from . import util
from . import errors

from . import weave_types as types
from . import artifact_wandb
from . import artifact_fs
from . import file_base
from . import file_util
from . import filesystem
from . import environment

WORKING_DIR_PREFIX = "__working__"


def local_artifact_dir() -> str:
    d = os.path.join(filesystem.get_filesystem_dir(), "local-artifacts")
    os.makedirs(d, exist_ok=True)
    return d


# From sdk/interface/artifacts.py
def md5_hash_file(path):
    hash_md5 = hashlib.md5()
    with open(path, "rb") as f:
        for chunk in iter(lambda: f.read(64 * 1024), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def md5_string(string: str) -> str:
    hash_md5 = hashlib.md5()
    hash_md5.update(string.encode())
    return hash_md5.hexdigest()


def local_artifact_exists(name: str, branch: str) -> bool:
    return os.path.exists(os.path.join(local_artifact_dir(), name, branch))


class LocalArtifactType(artifact_fs.FilesystemArtifactType):
    def save_instance(self, obj, artifact, name) -> "LocalArtifactRef":
        return LocalArtifactRef(obj, None)

    # No load_instance because we're returning a ref from save_instance.
    # Weave handles loading from that Ref automatically.


RESERVED_METADATA_KEYS = ["branch_point", "created_at", "previous_commit_uri"]


class LocalArtifact(artifact_fs.FilesystemArtifact):
    _existing_dirs: list[str]
    _original_uri: typing.Optional[str]
    _metadata: dict[str, typing.Any]

    def __init__(self, name: str, version: typing.Optional[str] = None):
        # LocalArtifacts are created frequently, sometimes in cases where
        # they will neither be read to or written to. The to_python path does
        # this, it creates an Artifact in case any of the objects in the tree
        # we're serializing are custom and therefore would need to write to
        # the artifact. But most times, there are no custom objects in the tree.
        #
        # So for performance, its important to not create the directory structure until
        # until we actually need to write to the artifact.
        if "/" in name or "\\" in name or ".." in name or ":" in name:
            raise ValueError('Artifact name cannot contain "/" or "\\" or ".." or ":"')
        self.name = name
        self._version = version
        self._branch: typing.Optional[str] = None
        self._root = os.path.join(local_artifact_dir(), name)
        self._original_uri = None
        self._path_handlers: dict[str, typing.Any] = {}
        self._existing_dirs = []
        self._metadata = {}
        self._setup_dirs()

    @classmethod
    def fork_from_uri(
        cls,
        source_uri: typing.Union[
            "WeaveLocalArtifactURI", artifact_wandb.WeaveWBArtifactURI
        ],
    ):
        art = cls(source_uri.name, source_uri.version)
        art._original_uri = str(source_uri)
        source_art = source_uri.to_ref().artifact
        if not isinstance(source_art, (LocalArtifact, artifact_wandb.WandbArtifact)):
            raise errors.WeaveInternalError(
                "Cannot fork from non-local artifact: %s" % source_art
            )
        art._version = source_art.version
        art._metadata = source_art.metadata.as_dict()
        return art

    # If an object has a _ref property, it will be used by ref_base._get_ref()
    @property
    def _ref(self) -> "LocalArtifactRef":
        if not self.is_saved:
            raise errors.WeaveInternalError("cannot get ref of an unsaved artifact")
        return LocalArtifactRef(self, None, None)

    def __repr__(self):
        return "<LocalArtifact(%s) %s %s>" % (id(self), self.name, self._version)

    def delete(self):
        shutil.rmtree(self._root)

    def rename(self, new_name: str):
        shutil.move(self._root, os.path.join(local_artifact_dir(), new_name))

    @property
    def is_saved(self) -> bool:
        return self._read_dirname is not None and self._version is not None

    @property
    def version(self):
        if not self.is_saved:
            raise errors.WeaveInternalError(
                "artifact must be saved before calling version!"
            )
        return self._version

    @property
    def created_at(self):
        return self.read_metadata()["created_at"]

    @property
    def branch(self):
        return self._branch

    @property
    def branch_point(self) -> typing.Optional[artifact_fs.BranchPointType]:
        return self.read_metadata().get("branch_point")

    def previous_uri(self) -> typing.Optional[str]:
        previous_commit_uri = self.read_metadata().get("previous_commit_uri")
        if previous_commit_uri is not None:
            return previous_commit_uri

        branch_point = self.branch_point
        if branch_point is not None:
            original_uri = branch_point["original_uri"]
            return original_uri
        return None

    def undo(self) -> typing.Optional[artifact_fs.FilesystemArtifact]:
        # Fetch the previous uri (could be from a local commit or a branch)
        previous_uri = self.previous_uri()
        if previous_uri is None:
            return None

        # Should we delete the current version?
        # self.delete()

        # Make sure we're undoing to an artifact uri
        original_uri_obj = uris.WeaveURI.parse(previous_uri)
        if not isinstance(
            original_uri_obj, (artifact_wandb.WeaveWBArtifactURI, WeaveLocalArtifactURI)
        ):
            raise errors.WeaveInternalError(
                "Cannot undo to non artifact uri: %s" % previous_uri
            )

        # Get the previous version
        previous_version = original_uri_obj.to_ref().artifact

        # If we are currently on a branch, and our previous version is local, then
        # we want to manually update the branch to point to the previous version.
        branch_name = self.branch
        if branch_name != None and isinstance(previous_version, LocalArtifact):
            commit_hash = previous_version.version
            if commit_hash != None and artifact_wandb.likely_commit_hash(commit_hash):
                previous_version._branch = branch_name
                link_name = os.path.join(previous_version._root, branch_name)
                target_name = os.path.join(previous_version._root, commit_hash)
                safe_os_symlink(target_name, link_name)

        # Return the previous version
        return previous_version

    def get_other_version(self, version: str) -> typing.Optional["LocalArtifact"]:
        if not local_artifact_exists(self.name, version):
            return None
        return LocalArtifact(self.name, version)

    def _setup_dirs(self):
        # NOTE: do not write to the filesystem here! This is called to construct
        # an artifact that may not exist or ever be read or written to.
        self._write_dirname = os.path.join(
            self._root, f"{WORKING_DIR_PREFIX}-{util.rand_string_n(12)}"
        )
        self._read_dirname = None
        if self._version:
            read_dirname = os.path.join(self._root, self._version)
            if not os.path.exists(read_dirname):
                # If it doesn't exist, assume the user has passed in a branch name
                # for version.
                self._branch = self._version
            else:
                self._read_dirname = os.path.join(self._root, self._version)

                # if this is a branch, set to the actual specific version it points to
                if os.path.islink(self._read_dirname):
                    self._branch = self._version
                    self._version = os.path.basename(
                        os.path.realpath(self._read_dirname)
                    )
                    self._read_dirname = os.path.join(self._root, self._version)
        if self._branch != None:
            self._original_uri = str(
                WeaveLocalArtifactURI(
                    self.name,
                    self._branch,
                )
            )

    def _get_read_path(self, path: str) -> pathlib.Path:
        read_dirname = pathlib.Path(self._read_dirname)
        full_path = read_dirname / path
        if not pathlib.Path(full_path).resolve().is_relative_to(read_dirname.resolve()):
            raise errors.WeaveAccessDeniedError()
        return full_path

    def _get_write_path(self, path: str) -> pathlib.Path:
        write_dirname = pathlib.Path(self._write_dirname)
        full_path = write_dirname / path
        if (
            not pathlib.Path(full_path)
            .resolve()
            .is_relative_to(write_dirname.resolve())
        ):
            raise errors.WeaveAccessDeniedError()
        return full_path

    def direct_url(self, name: str) -> str:
        art_path = self.path(name)
        local_path = os.path.abspath(art_path)
        return f"{environment.weave_server_url()}/__weave/file{local_path}"

    def path(self, name: str) -> str:
        return str(self._get_read_path(name))

    @property
    def initial_uri_obj(self) -> uris.WeaveURI:
        version = self._branch or self._version
        if version is None:
            raise errors.WeaveInternalError("Cannot get uri for unsaved artifact!")
        return WeaveLocalArtifactURI(
            self.name,
            version,
        )

    @property
    def uri_obj(self) -> uris.WeaveURI:
        version = self._version
        if version is None:
            raise errors.WeaveInternalError("Cannot get uri for unsaved artifact!")
        return WeaveLocalArtifactURI(
            self.name,
            version,
        )

    def path_uri(self, path) -> uris.WeaveURI:
        version = self._version
        if version is None:
            raise errors.WeaveInternalError("Cannot get uri for unsaved artifact!")
        return WeaveLocalArtifactURI(
            self.name,
            self._version,
            path,
        )

    def _makedir(self, dirname: str):
        # Keep track of directories we've already created so we don't
        # create them multiple times, makedir is expensive if you call
        # it a million times, especially on a network file store!
        if dirname not in self._existing_dirs:
            os.makedirs(dirname, exist_ok=True)
            self._existing_dirs.append(dirname)

    @contextlib.contextmanager
    def new_file(self, path, binary=False):
        full_path = os.path.join(self._write_dirname, path)
        self._makedir(os.path.dirname(full_path))
        mode = "w"
        if binary:
            mode = "wb"
        f = file_util.safe_open(full_path, mode)
        yield f
        f.close()

    @contextlib.contextmanager
    def new_dir(self, path):
        full_path = self._get_write_path(path)
        self._makedir(full_path)
        os.makedirs(full_path, exist_ok=True)
        yield full_path

    @contextlib.contextmanager
    def open(self, path, binary=False):
        mode = "r"
        if binary:
            mode = "rb"
        f = file_util.safe_open(os.path.join(self._read_dirname, path), mode)
        yield f
        f.close()

    def get_path_handler(self, path, handler_constructor):
        handler = self._path_handlers.get(path)
        if handler is None:
            handler = handler_constructor(self, path)
            self._path_handlers[path] = handler
        return handler

    def read_metadata(self):
        if not self._read_dirname:
            return {}
        with file_util.safe_open(
            os.path.join(self._read_dirname, ".artifact-version.json")
        ) as f:
            obj = json.load(f)
            obj["created_at"] = datetime.fromisoformat(obj["created_at"])
            return obj

    def write_metadata(self, dirname, metadata):
        self._makedir(dirname)
        with file_util.safe_open(
            os.path.join(dirname, ".artifact-version.json"), "w"
        ) as f:
            json.dump({"created_at": datetime.now().isoformat(), **metadata}, f)

    def save(self, branch=None):
        for handler in self._path_handlers.values():
            handler.close()
        self._path_handlers = {}
        manifest = {}
        if self._read_dirname:
            for dirpath, dnames, fnames in os.walk(self._read_dirname):
                for f in fnames:
                    if f != ".artifact-version.json":
                        full_path = os.path.join(dirpath, f)
                        manifest[f] = md5_hash_file(full_path)
        for dirpath, dnames, fnames in os.walk(self._write_dirname):
            for f in fnames:
                if f != ".artifact-version.json":
                    full_path = os.path.join(dirpath, f)
                    manifest[f] = md5_hash_file(full_path)
        commit_hash = md5_string(json.dumps(manifest, sort_keys=True, indent=2))[
            : artifact_wandb.WANDB_COMMIT_HASH_LENGTH
        ]

        new_dirname = os.path.join(self._root, commit_hash)

        if not self._read_dirname:
            # we're not read-modify-writing an existing version, so
            # just rename the write dir
            try:
                os.rename(self._write_dirname, new_dirname)
            except OSError:
                # Someone already created this version.
                if os.path.exists(self._write_dirname):
                    shutil.rmtree(self._write_dirname)
        else:
            # read-modify-write of existing version, so copy existing
            # files into new dir first, then overwrite with new files

            # using a working directory so we can atomic rename at end
            tmpdir = os.path.join(self._root, "tmpwritedir-%s" % util.rand_string_n(12))
            os.makedirs(tmpdir, exist_ok=True)
            if self._read_dirname:
                for path in os.listdir(self._read_dirname):
                    src_path = os.path.join(self._read_dirname, path)
                    target_path = os.path.join(tmpdir, path)
                    if os.path.isdir(src_path):
                        shutil.copytree(src_path, target_path)
                    else:
                        shutil.copyfile(src_path, target_path)
            if os.path.exists(self._write_dirname):
                for path in os.listdir(self._write_dirname):
                    src_path = os.path.join(self._write_dirname, path)
                    target_path = os.path.join(tmpdir, path)
                    if os.path.isdir(src_path):
                        shutil.copytree(src_path, target_path, dirs_exist_ok=True)
                    else:
                        shutil.copyfile(src_path, target_path)
            try:
                os.rename(tmpdir, new_dirname)
            except OSError:
                shutil.rmtree(tmpdir)
            if os.path.exists(self._write_dirname):
                shutil.rmtree(self._write_dirname)

        if branch is None:
            branch = self._branch

        metadata = {**self.metadata.as_dict()}
        # I don't think we need to delete this, but doing so
        # to be safe
        last_previous_uri = None
        for key in RESERVED_METADATA_KEYS:
            if key in metadata:
                if key == "previous_commit_uri":
                    last_previous_uri = metadata[key]
                del metadata[key]
        if self._original_uri != None and (
            branch != self._branch or self._original_uri.startswith("wandb-artifact://")
        ):
            # new branch
            metadata["branch_point"] = {
                "original_uri": self._original_uri,
                "commit": self._version,
                "n_commits": 1,
            }
        else:
            # same branch, same branch point
            if self.branch_point:
                metadata["branch_point"] = self.branch_point
                metadata["branch_point"]["n_commits"] += 1
        if (
            self.is_saved
            and self.version
            and artifact_wandb.likely_commit_hash(self.version)
            and self.version != commit_hash
        ):
            metadata["previous_commit_uri"] = str(self.uri_obj)
        elif last_previous_uri != None:
            # TODO: Refactor this to be more sane - we should be explicit about
            # what metadata carries over and protect against duplicate saves
            metadata["previous_commit_uri"] = last_previous_uri
        self.write_metadata(new_dirname, metadata)

        self._version = commit_hash
        self._setup_dirs()

        def make_link(dirname: str):
            link_name = os.path.join(self._root, dirname)
            safe_os_symlink(commit_hash, link_name)

        if branch is not None:
            self._branch = branch
            make_link(branch)
        else:
            make_link("latest")

    def _path_info(
        self, path: str
    ) -> typing.Optional[
        typing.Union[
            "artifact_fs.FilesystemArtifactFile", "artifact_fs.FilesystemArtifactDir"
        ]
    ]:
        read_dirname = pathlib.Path(self._read_dirname)
        local_path = self._get_read_path(path)
        if local_path.is_file():
            return artifact_fs.FilesystemArtifactFile(self, path)
        elif local_path.is_dir():
            sub_files = {}
            sub_dirs = {}
            for sub_path in local_path.iterdir():
                relpath = str(sub_path.relative_to(read_dirname))
                if relpath == ".artifact-version.json":
                    continue
                if sub_path.is_file():
                    sub_files[sub_path.name] = artifact_fs.FilesystemArtifactFile(
                        self, relpath
                    )
                else:
                    sub_dirs[sub_path.name] = file_base.SubDir(
                        relpath,
                        25,  # TODO: size
                        {},
                        {},
                    )
            return artifact_fs.FilesystemArtifactDir(
                self, path, 14, sub_dirs, sub_files
            )
        else:
            return None

    @property
    def metadata(self) -> artifact_fs.ArtifactMetadata:
        read_metadata = {
            k: v
            for k, v in self.read_metadata().items()
            if k not in RESERVED_METADATA_KEYS
        }
        return artifact_fs.ArtifactMetadata(self._metadata, read_metadata)


def safe_os_symlink(target_path: str, symlink_path: str):
    # Example of one of many races here
    # ensure tempdir root exists
    tmpdir_root = pathlib.Path(os.path.join(local_artifact_dir(), "tmp"))
    tmpdir_root.mkdir(exist_ok=True)
    tmp_path_name = f"tmplink-{util.rand_string_n(12)}"
    with tempfile.TemporaryDirectory(dir=tmpdir_root) as d:
        temp_path = os.path.join(d, tmp_path_name)
        os.symlink(target_path, temp_path)
        os.rename(temp_path, symlink_path)


LocalArtifactType.instance_classes = LocalArtifact


class LocalArtifactRef(artifact_fs.FilesystemArtifactRef):
    artifact: LocalArtifact

    def versions(self) -> list[artifact_fs.FilesystemArtifactRef]:
        artifact_path = os.path.join(local_artifact_dir(), self.artifact.name)
        versions = []
        for version_name in os.listdir(artifact_path):
            if (
                not os.path.islink(os.path.join(artifact_path, version_name))
                and not version_name.startswith(WORKING_DIR_PREFIX)
                and not version_name.startswith(".")
            ):
                # This is ass-backward, have to get the full object to just
                # get the ref.
                # TODO
                art = self.artifact.get_other_version(version_name)
                if art is None:
                    raise errors.WeaveInternalError(
                        "Could not get other version: %s %s"
                        % (self.artifact, version_name)
                    )
                ref = LocalArtifactRef(art, path="obj")
                # obj = uri.get()
                # ref = get_ref(obj)
                versions.append(ref)
        return sorted(versions, key=lambda v: v.artifact.created_at)

    @classmethod
    def from_uri(cls, uri: uris.WeaveURI) -> "LocalArtifactRef":
        if not isinstance(uri, WeaveLocalArtifactURI):
            raise errors.WeaveInternalError(
                f"Invalid URI class passed to WandbLocalArtifactRef.from_uri: {type(uri)}"
            )
        return cls(
            LocalArtifact(uri.name, uri.version),
            path=uri.path,
            obj=None,
            extra=uri.extra,
        )


types.LocalArtifactRefType.instance_class = LocalArtifactRef
types.LocalArtifactRefType.instance_classes = LocalArtifactRef

LocalArtifact.RefClass = LocalArtifactRef


@dataclasses.dataclass
class WeaveLocalArtifactURI(uris.WeaveURI):
    SCHEME = "local-artifact"
    path: typing.Optional[str] = None
    extra: typing.Optional[list[str]] = None

    @classmethod
    def from_parsed_uri(
        cls,
        uri: str,
        schema: str,
        netloc: str,
        path: str,
        params: str,
        query: dict[str, list[str]],
        fragment: str,
    ):
        parts = path.strip("/").split("/")
        name, version = parts[0].split(":", 1)
        file_path: typing.Optional[str] = None
        if len(parts) > 1:
            file_path = "/".join(parts[1:])
        extra: typing.Optional[list[str]] = None
        if fragment:
            extra = fragment.split("/")
        return cls(name, version, file_path, extra)

    def __str__(self) -> str:
        uri = f"{self.SCHEME}:///{self.name}:{self.version}"
        if self.path:
            uri += f"/{self.path}"
        if self.extra:
            uri += f"#{'/'.join(self.extra)}"
        return uri

    def to_ref(self) -> LocalArtifactRef:
        return LocalArtifactRef.from_uri(self)


def get_local_version_ref(name: str, version: str) -> typing.Optional[LocalArtifactRef]:
    # TODO: Watch out, this is a major race!
    #   - We need to eliminate this race or allow duplicate objectcs in parallel
    #     and then resolve later.
    #   - This is especially a problem for creating Runs and async Runs. We may
    #     accidentally launch parallel runs with the same run ID!
    if not local_artifact_exists(name, version):
        return None
    art = LocalArtifact(name, version)
    return LocalArtifactRef(art, path="obj")


# Should probably be "get_version_object()"
def get_local_version(name: str, version: str) -> typing.Any:
    ref = get_local_version_ref(name, version)
    if ref is None:
        return None
    return ref.get()



import dataclasses
import typing

import weave
from weave import weave_internal
from .. import panel
from .. import graph
from . import table_state


@weave.type()
class QueryDimsConfig:
    text: str


EditorType = typing.TypeVar("EditorType")


@weave.type()
class QueryCondition:
    expression: weave.Node[typing.Any] = dataclasses.field(
        default_factory=lambda: weave.graph.VoidNode()
    )
    editor: EditorType = dataclasses.field(default_factory=lambda: graph.VoidNode())  # type: ignore


@weave.type()
class QueryConfig:
    tableState: table_state.TableState
    dims: QueryDimsConfig
    conditions: list[QueryCondition] = dataclasses.field(default_factory=list)
    pinnedRows: dict[str, list[int]] = dataclasses.field(default_factory=dict)


@weave.type()
class Query(panel.Panel):
    id = "Query"
    input_node: weave.Node[list[typing.Any]]
    config: typing.Optional[QueryConfig] = dataclasses.field(
        default_factory=lambda: None
    )

    def __init__(self, input_node, vars=None, config=None, **options):
        super().__init__(input_node=input_node, vars=vars)
        self.config = config
        if self.config is None:
            table = table_state.TableState(self.input_node)
            self.config = QueryConfig(
                tableState=table,
                dims=QueryDimsConfig(
                    text=table.add_column(lambda row: graph.VoidNode(), name="text"),
                ),
                pinnedRows={"": [0]},
            )
        if "conditions" in options:
            conds = [
                cond(weave_internal.make_var_node(self.input_node.type, "queryInput"))
                for cond in options["conditions"]
            ]
            self.config.conditions = conds

    @weave.op(hidden=True)
    def selected_refine(self) -> weave.types.Type:
        return self.input_node.type

    @weave.op(
        output_type=lambda inputs: inputs["self"].input_node.output_type,
        refine_output_type=selected_refine,
    )
    def selected(self):
        # What a shame we have to execute the table :(

        # Apply Filters
        table_node = self.input_node
        if (
            self.config
            and self.config.tableState.preFilterFunction is not None
            and self.config.tableState.preFilterFunction.type != weave.types.Invalid()
        ):
            table_node = weave.ops.List.filter(
                table_node,
                lambda row: weave_internal.call_fn(
                    self.config.tableState.preFilterFunction, {"row": row}
                ),
            )
        return table_node



"""
There are a number of panels defined in the JS implementation that are not
defined in the Python implementation. A dev can determine the total list of
panels by inspecting the panels in PanelRegistry.tsx. This is tested in
test_panel_coverage.py.

It is expected that panels with specific configuration options may be extracted
from this list and defined manually.
"""

import dataclasses
import typing
import weave


@dataclasses.dataclass
class LPanel:
    panel_id: str
    typename_override: typing.Optional[str] = None


panels = [
    LPanel("barchart", "PanelBarchart"),
    LPanel("web-viz", "PanelWebViz"),
    LPanel("video-file", "PanelVideoFile"),
    LPanel("model-file", "PanelModelFile"),
    LPanel("id-count", "PanelIdCount"),
    LPanel("wb_trace_tree-traceDebugger"),
    LPanel("link", "PanelLink"),
    LPanel("run-overview", "PanelRunOverview"),
    LPanel("none", "PanelNone"),
    LPanel("artifactVersionAliases", "PanelArtifactVersionAliases"),
    LPanel("netron", "PanelNetron"),
    LPanel("object", "PanelObject"),
    LPanel("audio-file", "PanelAudioFile"),
    LPanel("string-histogram", "PanelStringHistogram"),
    LPanel("rawimage", "PanelRawimage"),
    LPanel("precomputed-histogram", "PanelPrecomputedHistogram"),
    LPanel("image-file-compare", "PanelImageFileCompare"),
    LPanel("molecule-file", "PanelMoleculeFile"),
    LPanel("multi-histogram", "PanelMultiHistogram"),
    LPanel("object3D-file", "PanelObject3DFile"),
    LPanel("run-color", "PanelRunColor"),
    LPanel("multi-string-histogram", "PanelMultiStringHistogram"),
    LPanel("dir", "PanelDir"),
    LPanel("id-compare-count", "PanelIdCompareCount"),
    LPanel("jupyter", "PanelJupyter"),
    LPanel("bokeh-file", "PanelBokehFile"),
    LPanel("ndarray", "PanelNdarray"),
    LPanel("id-compare", "PanelIdCompare"),
    LPanel("unknown", "PanelUnknown"),
    LPanel("image-file", "PanelImageFile"),
    LPanel("project-overview", "PanelProjectOverview"),
    LPanel("textdiff", "PanelTextdiff"),
    LPanel("type", "PanelType"),
    LPanel("text", "PanelText"),
    LPanel("string-compare", "PanelStringCompare"),
    LPanel("debug-expression-graph", "PanelDebugExpressionGraph"),
    LPanel("tracer", "PanelTracer"),
    LPanel("projection.plot", "projection.plot"),
    LPanel("RootBrowser", "RootBrowser"),
    LPanel("row.runs-table.row.projection.plot", "row.runs-table.row.projection.plot"),
    LPanel("maybe.plot", "maybe.plot"),
    LPanel("maybe.wb_trace_tree-traceViewer", "maybe.wb_trace_tree-traceViewer"),
]

did_define = False


def define_panel(p: LPanel):
    class DummyClass(weave.panel.Panel):
        id = p.panel_id

    name = p.typename_override or p.panel_id
    if weave.types.type_name_to_type(name) is not None:
        raise RuntimeError(f"Panel {name} already defined")
    return weave.type(__override_name=name)(DummyClass)  # type: ignore


if not did_define:
    for p in panels:
        define_panel(p)
        if p.typename_override is not None:
            define_panel(LPanel("maybe." + p.panel_id, "maybe." + p.panel_id))

did_define = True



import typing
from ..api import op
from ..compile_domain import InputProvider, wb_gql_op_plugin
from .. import weave_types
from inspect import signature, Parameter
from . import wb_domain_types
import hashlib
from .. import errors

"""
This file contains utilities for constructing GQL ops (used by all the ops in
`ops_domain`). There are explicitly 4 functions:
    - `gql_prop_op`: used for getting properties of GQL objects
    - `gql_root_op`: used to start a query (much match to a query on the `Query`
      type)
    - `gql_direct_edge_op`: used to get a direct edge of a GQL object leading to
      another GQL object
        - Has an additional `is_many` arg to determine if the result is a single
          or many objects
    - `gql_connection_op`: used to get a connection of GQL objects (the standard
      edges/nodes pattern)

All but the prop_op have the ability to specify additional inputs and how to map
such inputs to a query param string.

Each of these leverage the underlying `wb_gql_op_plugin` to create an op. You
are always free to directly use `wb_gql_op_plugin` if you need to write custom
logic - these are just wrappers since these cases are so common. 

Please see `project_ops.py` for examples of all the above cases.
"""


def _make_alias(*args: str, prefix: str = "alias") -> str:
    inputs = "_".join([str(arg) for arg in args])
    digest = hashlib.md5(inputs.encode()).hexdigest()
    return f"{prefix}_{digest}"


def gql_prop_op(
    op_name: str,
    input_type: weave_types.Type,
    prop_name: str,
    output_type: weave_types.Type,
):
    first_arg_name = input_type.name

    def gql_property_getter_op_fn(**inputs):
        if isinstance(output_type, weave_types.Timestamp):
            return output_type.from_isostring(inputs[first_arg_name].gql[prop_name])
        return inputs[first_arg_name].gql[prop_name]

    sig = signature(gql_property_getter_op_fn)
    params = [Parameter(first_arg_name, Parameter.POSITIONAL_OR_KEYWORD)]
    sig = sig.replace(parameters=tuple(params))
    gql_property_getter_op_fn.__signature__ = sig  # type: ignore
    gql_property_getter_op_fn.sig = sig  # type: ignore

    return op(
        name=op_name,
        plugins=wb_gql_op_plugin(
            lambda inputs, inner: prop_name,
        ),
        input_type={first_arg_name: input_type},
        output_type=output_type,
    )(gql_property_getter_op_fn)


def _get_required_fragment(output_type: weave_types.Type):
    return (
        output_type.instance_class.REQUIRED_FRAGMENT  # type: ignore
        if output_type.instance_class is not None
        and hasattr(output_type.instance_class, "REQUIRED_FRAGMENT")
        else ""
    )


def gql_root_op(
    op_name: str,
    prop_name: str,
    output_type: weave_types.Type,
    additional_inputs_types: typing.Optional[dict[str, weave_types.Type]] = None,
    param_str_fn: typing.Optional[typing.Callable[[InputProvider], str]] = None,
):
    return gql_direct_edge_op(
        op_name,
        None,
        prop_name,
        output_type,
        additional_inputs_types,
        param_str_fn,
    )


def gql_direct_edge_op(
    op_name: str,
    input_type: typing.Optional[weave_types.Type],
    prop_name: str,
    output_type: weave_types.Type,
    additional_inputs_types: typing.Optional[dict[str, weave_types.Type]] = None,
    param_str_fn: typing.Optional[typing.Callable[[InputProvider], str]] = None,
    is_many: bool = False,
):
    is_root = input_type is None
    first_arg_name = "gql_obj" if input_type is None else input_type.name
    if not output_type.instance_class or isinstance(
        output_type.instance_class, wb_domain_types.GQLTypeMixin
    ):
        raise ValueError(
            f"output_type must be a GQLTypeMixin, got {output_type} instead"
        )

    def query_fn(inputs, inner):
        alias = ""
        param_str = ""
        if param_str_fn:
            param_str = param_str_fn(inputs)
            alias = f"{_make_alias(param_str, prefix=prop_name)}:"
            param_str = f"({param_str})"
        return f"""
            {alias} {prop_name}{param_str} {{
                {_get_required_fragment(output_type)}
                {inner}
            }}
        """

    additional_inputs_types = additional_inputs_types or {}
    if is_root:

        def gql_relationship_getter_op(**inputs):
            raise errors.WeaveGQLCompileError(
                f"{op_name} should not be executed directly. If you see this error, it is a bug in the Weave compiler."
            )

    else:

        def gql_relationship_getter_op(**inputs):
            gql_obj = inputs[first_arg_name]
            additional_inputs = {
                key: value for key, value in inputs.items() if key != first_arg_name
            }
            name = prop_name
            if param_str_fn:
                param_str = param_str_fn(InputProvider(additional_inputs))
                name = _make_alias(param_str, prefix=prop_name)
            if is_many:
                return [
                    output_type.instance_class.from_gql(item)
                    for item in gql_obj.gql[name]
                ]
            if gql_obj.gql == wb_domain_types.UntypedOpaqueDict.from_json_dict(None):
                return None
            gql_val = gql_obj.gql.get(name)
            if gql_val is None:
                return None
            return output_type.instance_class.from_gql(gql_val)

    sig = signature(gql_relationship_getter_op)
    base_params = (
        [Parameter(first_arg_name, Parameter.POSITIONAL_OR_KEYWORD)]
        if not is_root
        else []
    )
    new_params = [
        *base_params,
        *[
            Parameter(key, Parameter.POSITIONAL_OR_KEYWORD)
            for key in additional_inputs_types.keys()
        ],
    ]
    sig = sig.replace(parameters=tuple(new_params))
    gql_relationship_getter_op.__signature__ = sig  # type: ignore
    gql_relationship_getter_op.sig = sig  # type: ignore

    base_input_type = {first_arg_name: input_type} if not is_root else {}
    _output_type = output_type
    if is_many:
        _output_type = weave_types.List(output_type)
    gql_relationship_getter_op = op(
        name=op_name,
        plugins=wb_gql_op_plugin(query_fn, is_root),
        input_type={**base_input_type, **additional_inputs_types},
        output_type=_output_type,
    )(gql_relationship_getter_op)

    return gql_relationship_getter_op


def gql_connection_op(
    op_name: str,
    input_type: weave_types.Type,
    prop_name: str,
    output_type: weave_types.Type,
    additional_inputs_types: typing.Optional[dict[str, weave_types.Type]] = None,
    param_str_fn: typing.Optional[typing.Callable[[InputProvider], str]] = None,
):
    first_arg_name = "gql_obj" if input_type is None else input_type.name
    if not output_type.instance_class or isinstance(
        output_type.instance_class, wb_domain_types.GQLTypeMixin
    ):
        raise ValueError(
            f"output_type must be a GQLTypeMixin, got {output_type} instead"
        )

    def query_fn(inputs, inner):
        alias = ""
        param_str = ""
        if param_str_fn:
            param_str = param_str_fn(inputs)
            alias = f"{_make_alias(param_str, prefix=prop_name)}:"
            param_str = f"({param_str})"
        return f"""
            {alias} {prop_name}{param_str} {{
                edges {{
                    node {{
                        {_get_required_fragment(output_type)}
                        {inner}
                    }}
                }}
            }}
        """

    additional_inputs_types = additional_inputs_types or {}

    def gql_connection_walker_op(**inputs):
        gql_obj = inputs[first_arg_name]
        additional_inputs = {
            key: value for key, value in inputs.items() if key != first_arg_name
        }
        name = prop_name
        if param_str_fn:
            param_str = param_str_fn(InputProvider(additional_inputs))
            name = _make_alias(param_str, prefix=prop_name)
        # If we have a None argument, return an empty list.
        if gql_obj.gql == wb_domain_types.UntypedOpaqueDict.from_json_dict(None):
            return []
        return [
            output_type.instance_class.from_gql(edge["node"])
            for edge in gql_obj.gql[name]["edges"]
        ]

    sig = signature(gql_connection_walker_op)
    new_params = [
        Parameter(first_arg_name, Parameter.POSITIONAL_OR_KEYWORD),
        *[
            Parameter(key, Parameter.POSITIONAL_OR_KEYWORD)
            for key in additional_inputs_types.keys()
        ],
    ]
    sig = sig.replace(parameters=tuple(new_params))
    gql_connection_walker_op.__signature__ = sig  # type: ignore
    gql_connection_walker_op.sig = sig  # type: ignore
    gql_connection_walker_op = op(
        name=op_name,
        plugins=wb_gql_op_plugin(query_fn),
        input_type={first_arg_name: input_type, **additional_inputs_types},
        output_type=weave_types.List(output_type),
    )(gql_connection_walker_op)

    return gql_connection_walker_op



from ..compile_domain import wb_gql_op_plugin
from ..api import op
from .. import weave_types as types
from . import wb_domain_types as wdt
from ..language_features.tagging.make_tag_getter_op import make_tag_getter_op
from .wandb_domain_gql import (
    gql_prop_op,
    gql_direct_edge_op,
    gql_connection_op,
    gql_root_op,
)

# Section 1/6: Tag Getters
# None

# Section 2/6: Root Ops
root_viewer = gql_root_op(
    "root-viewer",
    "viewer",
    wdt.UserType,
)

root_user = gql_root_op(
    "root-user",
    "user",
    wdt.UserType,
    {
        "userName": types.String(),
    },
    lambda inputs: f'userName: {inputs["userName"]}',
)

# Section 3/6: Attribute Getters
gql_prop_op(
    "user-username",
    wdt.UserType,
    "username",
    types.String(),
)

gql_prop_op(
    "user-name",
    wdt.UserType,
    "name",
    types.String(),
)

# Section 4/6: Direct Relationship Ops
# None

# Section 5/6: Connection Ops
gql_connection_op(
    "user-entities",
    wdt.UserType,
    "teams",
    wdt.EntityType,
    {},
    lambda inputs: f"first: 100",
)

# Section 6/6: Non Standard Business Logic Ops
@op(name="user-link")
def link(project: wdt.User) -> wdt.Link:
    return wdt.Link(project.gql["name"], project.gql["name"])



import collections
import textwrap
import json
import inspect
import types as py_types
import typing
import os
import sys
import ast

from . import artifact_local
from . import op_def
from . import errors
from . import context_state
from . import weave_types as types
from . import registry_mem
from . import errors

from . import infer_types


def type_code(type_):
    if isinstance(type_, py_types.GenericAlias) or isinstance(
        type_, typing._GenericAlias  # type: ignore
    ):
        args = ", ".join(type_code(t) for t in type_.__args__)
        if type_.__origin__ == list or type_.__origin__ == collections.abc.Sequence:
            return f"list[{args}]"
        elif type_.__origin__ == dict:
            return f"dict[{args}]"
        elif type_.__origin__ == typing.Union:
            return f"typing.Union[{args}]"
        else:
            return f"{type_.__origin__}[{args}]"
    else:
        return type_.__name__


def generate_referenced_type_code(type_):
    # Given a function that may have type annotations, generate non-redundant
    # code that declares any referenced types and their referenced types and
    # so on.

    # Absolutely horrible and hacky. This is not recursive in a tree/graph
    # way, its linear, so it'll only produce one TypedDict if there are many.
    # Using this to get the versioned object notebook working.
    if infer_types.is_typed_dict_like(type_):
        result = f"class {type_.__name__}(typing.TypedDict):\n"
        for k in type_.__annotations__:
            result += f"    {k}: {type_code(type_.__annotations__[k])}\n"
        return result
    elif isinstance(type_, py_types.GenericAlias) or isinstance(
        type_, typing._GenericAlias  # type: ignore
    ):
        return generate_referenced_type_code(type_.__args__[0])


def get_code_deps(fn, decl_locals):
    # Pretty horrible and hacky POC.
    # Tries to pull in other functions and modules referenced in the function
    # body. Generates a lot of repetition and not general. I just made it
    # work for specific cases that exist in the example notebooks currently.
    source = inspect.getsource(fn)
    try:
        parsed = ast.parse(source)
    except IndentationError:
        return ""
    import_code = ""
    for node in ast.walk(parsed):
        if isinstance(node, ast.Name) and isinstance(node.ctx, ast.Load):
            # A really bad way to check if a variable reference is for
            # a module!
            local_val = decl_locals.get(node.id)
            if isinstance(local_val, py_types.ModuleType):
                import_code += f"import {local_val.__name__} as {node.id}\n"
            elif isinstance(local_val, py_types.FunctionType):
                import_code += get_code_deps(local_val, decl_locals)
                import_code += inspect.getsource(local_val)
    return import_code


class OpDefType(types.Type):
    instance_class = op_def.OpDef
    instance_classes = op_def.OpDef

    def save_instance(self, obj: op_def.OpDef, artifact, name):
        if obj.is_builtin:
            with artifact.new_file(f"{name}.json") as f:
                json.dump({"name": obj.name}, f)
        else:
            code = "import typing\nimport weave\n"

            # Try to figure out module imports from the function body
            # (in a real hacky way as a POC)
            code += get_code_deps(obj.raw_resolve_fn, obj._decl_locals)

            # Create TypedDict types for referenced TypedDicts
            resolve_annotations = obj.raw_resolve_fn.__annotations__
            for k, type_ in resolve_annotations.items():
                gen_type_code = generate_referenced_type_code(type_)
                if gen_type_code is not None:
                    code += gen_type_code

            code += textwrap.dedent(inspect.getsource(obj.raw_resolve_fn))
            with artifact.new_file(f"{name}.py") as f:
                f.write(code)

    def load_instance(cls, artifact, name, extra=None):
        try:
            with artifact.open(f"{name}.json") as f:
                op_spec = json.load(f)

            return registry_mem.memory_registry._ops[op_spec["name"]]
        except FileNotFoundError:
            pass

        path_with_ext = os.path.relpath(
            artifact.path(f"{name}.py"), start=artifact_local.local_artifact_dir()
        )
        # remove the .py extension
        path = os.path.splitext(path_with_ext)[0]
        # convert filename into module path
        parts = path.split("/")
        module_path = ".".join(parts)

        sys.path.insert(0, artifact_local.local_artifact_dir())
        with context_state.loading_op_location(artifact.uri_obj):
            # This has a side effect of registering the op
            mod = __import__(module_path)
        sys.path.pop(0)
        # We justed imported e.g. 'op-number-add.xaybjaa._obj'. Navigate from
        # mod down to _obj.
        for part in parts[1:]:
            mod = getattr(mod, part)

        op_defs = inspect.getmembers(mod, op_def.is_op_def)
        if len(op_defs) != 1:
            raise errors.WeaveInternalError(
                "Unexpected Weave module saved in: %s" % path
            )
        _, od = op_defs[0]
        return od


def fully_qualified_opname(wrap_fn):
    op_module_file = os.path.abspath(inspect.getfile(wrap_fn))
    if op_module_file.endswith(".py"):
        op_module_file = op_module_file[:-3]
    elif op_module_file.endswith(".pyc"):
        op_module_file = op_module_file[:-4]
    return "file://" + op_module_file + "." + wrap_fn.__name__



import pyarrow.compute as pc
import pyarrow as pa

from ..decorator_op import op
from ..decorator_arrow_op import arrow_op
from .. import weave_types as types
from . import util
from .list_ import ArrowWeaveList, ArrowWeaveListType

unary_input_type = {
    "self": ArrowWeaveListType(types.Boolean()),
}
binary_input_type = {
    "self": ArrowWeaveListType(types.Boolean()),
    "other": types.UnionType(types.Boolean(), ArrowWeaveListType(types.Boolean())),
}

nullable_binary_input_type = {
    "self": ArrowWeaveListType(types.optional(types.Boolean())),
    "other": types.UnionType(
        types.optional(types.Boolean()),
        ArrowWeaveListType(types.optional(types.Boolean())),
    ),
}


null_consuming_binary_input_type_right = {
    "self": types.optional(types.Boolean()),
    "other": ArrowWeaveListType(types.optional(types.Boolean())),
}


self_type_output_type_fn = lambda input_types: input_types["self"]


@arrow_op(
    name="ArrowWeaveListBoolean-equal",
    input_type=nullable_binary_input_type,
    output_type=ArrowWeaveListType(types.Boolean()),
)
def bool_equal(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    result = util.equal(self._arrow_data, other)
    return ArrowWeaveList(result, types.Boolean(), self._artifact)


@arrow_op(
    name="ArrowWeaveListBoolean-notEqual",
    input_type=nullable_binary_input_type,
    output_type=ArrowWeaveListType(types.Boolean()),
)
def bool_not_equal(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    result = util.not_equal(self._arrow_data, other)
    return ArrowWeaveList(result, types.Boolean(), self._artifact)


@arrow_op(
    name="ArrowWeaveListBoolean-and",
    input_type=binary_input_type,
    output_type=self_type_output_type_fn,
)
def bool_and(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    return ArrowWeaveList(
        pc.and_(self._arrow_data, other), types.Boolean(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListBoolean-or",
    input_type=binary_input_type,
    output_type=self_type_output_type_fn,
)
def bool_or(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    return ArrowWeaveList(
        pc.or_(self._arrow_data, other), types.Boolean(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListBoolean-not",
    input_type=unary_input_type,
    output_type=self_type_output_type_fn,
)
def bool_not(self):
    return ArrowWeaveList(pc.invert(self._arrow_data), types.Boolean(), self._artifact)



import typing


def parse_local_ref_str(s: str) -> typing.Tuple[str, typing.Optional[list[str]]]:
    if "#" not in s:
        return s, None
    path, extra = s.split("#", 1)
    return path, extra.split("/")



# Top-Level Namespace: weave.monitoring

# TODO: Notebook Walkthrough

from ..wandb_interface.wandb_stream_table import StreamTable
from .monitor_decorator import monitor


__all__ = ["StreamTable", "monitor"]



from . import api as weave
from PIL import Image


@weave.type()
class Point2D:
    x: float
    y: float

    @weave.op()
    def get_x(self) -> float:
        return self.x


@weave.type()
class Size2D:
    w: float
    h: float


@weave.type()
class BoundingBox2D:
    top_left: Point2D
    size: Size2D

    @weave.op()
    def center(self) -> Point2D:
        return Point2D(
            self.top_left.x + self.size.w / 2, self.top_left.y + self.size.h / 2
        )


@weave.type()
class ImageWithBoxes:
    im: Image.Image
    boxes: list[BoundingBox2D]

    @weave.op()
    def get_boxes(self) -> list[BoundingBox2D]:
        return self.boxes



from ..decorators import op
from .. import weave_types as types


def get_const_union_vals(type_, of_type):
    if isinstance(type_, types.Const):
        if type_.val_type != of_type:
            return None
        return [type_.val]
    if not isinstance(type_, types.UnionType):
        return None
    vals = []
    for m in type_.members:
        if not isinstance(m, types.Const) or not m.val_type == of_type:
            return None
        vals.append(m.val)
    return vals


def union_output_type(input_type):
    s1_type = input_type["s1"].object_type
    s2_type = input_type["s2"].object_type
    vals1 = get_const_union_vals(s1_type, types.String())
    vals2 = get_const_union_vals(s2_type, types.String())
    if vals1 and vals2:
        union_vals = set(vals1).union(vals2)
        return types.List(
            types.UnionType(*(types.Const(types.String(), v) for v in union_vals))
        )
    return types.List(types.String())


# TODO: generic
@op(output_type=union_output_type)
def union(s1: list[str], s2: list[str]):
    return list(set(s1).union(set(s2)))


def difference_output_type(input_type):
    s1_type = input_type["s1"].object_type
    s2_type = input_type["s2"].object_type
    vals1 = get_const_union_vals(s1_type, types.String())
    vals2 = get_const_union_vals(s2_type, types.String())
    if vals1 and vals2:
        union_vals = set(vals1).difference(vals2)
        return types.List(
            types.union(*(types.Const(types.String(), v) for v in union_vals))
        )
    return types.List(types.String())


@op(output_type=difference_output_type)
def difference(s1: list[str], s2: list[str]):
    return list(set(s1).difference(set(s2)))



import analytics
import subprocess

from . import context_state

analytics.write_key = "uJ8vZgKqTBVH6ZdhD4GZGZYsR7ucfJmb"


def whoami():
    return subprocess.check_output("whoami").decode().strip()


identify_called = False


def _identify():
    global identify_called
    if not identify_called:
        who = whoami()
        analytics.identify(who, {"whoami": who})
    identify_called = True


def track(action: str, info=None):
    if not context_state.analytics_enabled():
        return
    _identify()
    if info is None:
        info = {}
    analytics.track(whoami(), action, info)


def use_called():
    track("called use")


def show_called(info=None):
    track("called show", info)



from sklearn.datasets import fetch_california_housing

import weave


# TODO: this should return a Weave type, not a raw dataframe
@weave.op(
    name="shap-ca_housing_dataset",
    render_info={"type": "function"},
    output_type=weave.ops.DataFrameType(weave.types.TypedDict({})),
)
def ca_housing_dataset(seed: int):
    housing = fetch_california_housing(as_frame=True)
    housingdf = housing.frame
    return housingdf



import os
from pathlib import Path
from setuptools import setup  # type: ignore[import]
from setuptools.command.build import build  # type: ignore[import]
from setuptools.command.editable_wheel import editable_wheel  # type: ignore[import]
from setuptools.command.sdist import sdist  # type: ignore[import]
import subprocess
import tarfile
import tempfile
from typing import Union
import urllib.request
from urllib.error import HTTPError

ROOT = Path(__file__).resolve().parent
SKIP_BUILD = os.environ.get("WEAVE_SKIP_BUILD", False)
IS_BUILT = (ROOT / "weave" / "frontend" / "assets").is_dir() or SKIP_BUILD
FORCE_BUILD = os.environ.get("WEAVE_FORCE_BUILD", False)


def check_build_deps() -> bool:
    have_yarn = False
    try:
        subprocess.run(["yarn", "--version"], capture_output=True)
        have_yarn = True
    except OSError:
        pass

    if not have_yarn:
        try:
            print("Attempting to install yarn...")
            subprocess.run(["npm", "install", "-g", "yarn"], capture_output=True)
        except OSError:
            raise RuntimeError(
                "You must have node v16+ (https://nodejs.org/en/download) installed to build weave."
            )
    return True


def build_frontend() -> None:
    check_build_deps()
    try:
        build_script = str(Path("weave", "frontend", "build.sh"))
        subprocess.run(["bash", build_script], cwd=ROOT)
    except OSError:
        raise RuntimeError("Failed to build frontend.")


def download_and_extract_tarball(
    url: str, extract_path: Union[Path, str] = "."
) -> None:
    file_name = os.path.basename(url)
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = os.path.join(temp_dir, file_name)
        try:
            # Download the tarball
            urllib.request.urlretrieve(url, temp_path)
        except urllib.error.URLError as e:
            raise RuntimeError(f"Couldn't download the tarball {url} due to error: {e}")

        try:
            # Extract the tarball
            if tarfile.is_tarfile(temp_path):
                with tarfile.open(temp_path, "r:gz") as tar:
                    tar.extractall(path=extract_path)
                # TODO: detect when the extracted assets are out of sync with git
            else:
                raise RuntimeError(f"{file_name} is not a tarball file.")
        except tarfile.TarError as e:
            raise RuntimeError(f"Couldn't extract {file_name} due to error: {e}")


def download_frontend() -> None:
    sha = open(ROOT / "weave" / "frontend" / "sha1.txt").read().strip()
    url = f"https://storage.googleapis.com/wandb-cdn-prod/weave/{sha}.tar.gz"
    try:
        download_and_extract_tarball(url, extract_path=ROOT / "weave")
    except HTTPError:
        print(f"Warning: Failed to download frontend for sha {sha}")


class Build(build):  # type: ignore
    def run(self) -> None:
        if FORCE_BUILD:
            build_frontend()
        elif not IS_BUILT:
            download_frontend()
        super().run()


class EditableWheel(editable_wheel):  # type: ignore
    def run(self) -> None:
        if FORCE_BUILD:
            build_frontend()
        elif not IS_BUILT:
            download_frontend()
        super().run()


class Sdist(sdist):  # type: ignore
    def run(self) -> None:
        if FORCE_BUILD:
            build_frontend()
        elif not IS_BUILT:
            download_frontend()
        super().run()


setup(cmdclass={"build": Build, "editable_wheel": EditableWheel, "sdist": Sdist})



import typing
from .. import weave_types as types
from .. import errors
from ..language_features.tagging import tag_store, tagged_value_type
from .. import box


def typeddict_pick_output_type(input_types):
    property_types = input_types["self"].property_types
    if not isinstance(input_types["key"], types.Const):
        return types.union(*property_types.values())
    key = input_types["key"].val
    output_type = _tag_aware_dict_or_list_type_for_path(
        input_types["self"], split_escaped_string(key)
    )
    return output_type


def tag_aware_dict_val_for_escaped_key(
    obj: dict, key: typing.Optional[str]
) -> typing.Any:
    if key == None:
        return None
    return _any_val_for_path(obj, split_escaped_string(key))


class MergeInputTypes(typing.TypedDict):
    self: typing.Union[types.TypedDict, types.UnionType, types.NoneType]
    other: types.TypedDict


def typeddict_merge_output_type(
    input_types: MergeInputTypes,
) -> typing.Union[types.TypedDict, types.UnionType, types.NoneType, types.UnknownType]:
    self = input_types["self"]
    other = input_types["other"]
    self_ok = types.UnionType(types.TypedDict({}), types.NoneType()).assign_type(self)
    if not self_ok or not isinstance(other, types.TypedDict):
        return types.UnknownType()

    # create property types without merging nested dictionary types (for now)
    if isinstance(self, types.TypedDict):
        property_types = {
            **self.property_types,
            **other.property_types,
        }

        return types.TypedDict(property_types=property_types)
    elif isinstance(self, types.UnionType):
        if not self.is_simple_nullable():
            return types.UnknownType()
        non_null_member = types.non_none(self)
        return types.UnionType(
            typeddict_merge_output_type(
                {"self": typing.cast(types.TypedDict, non_null_member), "other": other}
            ),
            types.NoneType(),
        )
    elif isinstance(self, types.NoneType):
        return types.NoneType()
    else:
        raise errors.WeaveTypeError(f"Unhandled types for dict merge: {self}, {other}")


# Re-implementation of helpers2.py
def unescape_dots(s: str) -> str:
    return s.replace("\\.", ".")


def split_escaped_string(s: typing.Optional[str]) -> list[str]:
    # splits a string on dots, but ignores dots that are escaped
    # e.g. "a.b.c" -> ["a", "b", "c"]
    # e.g. "a\\.b.c" -> ["a.b", "c"]
    if s is None:
        return []
    placeholder = "___DOT___"
    s = s.replace("\\.", placeholder)
    parts = s.split(".")
    parts = [p.replace(placeholder, ".") for p in parts]
    return parts


def _val_tag_wrapper(val: typing.Any) -> typing.Callable[[typing.Any], typing.Any]:
    tags = None
    if tag_store.is_tagged(val):
        tags = tag_store.get_tags(val)

    def wrapper(res: typing.Any) -> typing.Any:
        if tags is not None:
            return tag_store.add_tags(box.box(res), tags)
        else:
            return res

    return wrapper


def _any_val_for_path(val: typing.Any, path: list[str]) -> typing.Any:
    tag_wrapper = _val_tag_wrapper(val)
    if len(path) == 0:
        return None
    key = path[0]
    next_path = path[1:]
    if isinstance(val, list) and key == "*":
        if len(next_path) == 0:
            return val
        else:
            return tag_wrapper([_any_val_for_path(item, next_path) for item in val])
    elif isinstance(val, dict):
        return _dict_val_for_path(val, path)
    else:
        return None


def _dict_val_for_path(val: dict, path: list[str]) -> typing.Any:
    tag_wrapper = _val_tag_wrapper(val)
    if len(path) == 0:
        return None
    key = path[0]
    next_path = path[1:]
    if key == "*":
        return tag_wrapper(
            {
                sub_key: _any_val_for_path(sub_val, next_path)
                for sub_key, sub_val in val.items()
            }
        )
    if key not in val:
        return None
    key_val = val.get(key)
    if len(next_path) == 0:
        res = key_val
    else:
        res = _any_val_for_path(key_val, next_path)
    return tag_wrapper(res)


def _type_tag_wrapper(
    type: types.Type,
) -> typing.Tuple[typing.Callable[[types.Type], types.Type], types.Type]:
    tags = None
    inner_type = type
    if isinstance(type, tagged_value_type.TaggedValueType):
        tags = type.tag
        inner_type = type.value

    def wrapper(res: typing.Any) -> typing.Any:
        if tags is not None:
            return tagged_value_type.TaggedValueType(tags, res)
        else:
            return res

    return wrapper, inner_type


def _tag_aware_dict_or_list_type_for_path(
    type: types.Type, path: list[str]
) -> types.Type:
    if isinstance(type, types.Const):
        type = type.val_type
    tag_wrapper, inner_type = _type_tag_wrapper(type)
    if len(path) == 0:
        return types.NoneType()
    if isinstance(inner_type, types.UnionType):
        return tag_wrapper(
            types.union(
                *[
                    _tag_aware_dict_or_list_type_for_path(mem_type, path)
                    for mem_type in inner_type.members
                ]
            )
        )
    if not (
        isinstance(inner_type, types.List) or isinstance(inner_type, types.TypedDict)
    ):
        return types.NoneType()
    key = path[0]
    next_path = path[1:]
    if isinstance(inner_type, types.List) and key == "*":
        if len(next_path) == 0:
            return type
        else:
            return tag_wrapper(
                types.List(
                    _tag_aware_dict_or_list_type_for_path(
                        inner_type.object_type, next_path
                    )
                )
            )
    elif isinstance(inner_type, types.TypedDict):
        return _dict_type_for_path(type, path)
    else:
        return types.NoneType()


def _dict_type_for_path(type: types.Type, path: list[str]) -> types.Type:
    tag_wrapper, inner_type = _type_tag_wrapper(type)
    if len(path) == 0:
        return types.NoneType()
    if not isinstance(inner_type, types.TypedDict):
        return types.NoneType()
    prop_types = inner_type.property_types
    key = path[0]
    next_path = path[1:]
    if key == "*":
        return tag_wrapper(
            types.TypedDict(
                {
                    sub_key: _tag_aware_dict_or_list_type_for_path(sub_val, next_path)
                    for sub_key, sub_val in prop_types.items()
                }
            )
        )
    if key not in prop_types:
        return types.NoneType()
    key_val = typing.cast(types.Type, prop_types.get(key))
    if len(next_path) == 0:
        res = key_val
    else:
        res = _tag_aware_dict_or_list_type_for_path(key_val, next_path)

    return tag_wrapper(res)



import typing
from ..api import op, weave_class
from .. import weave_types as types


@weave_class(weave_type=types.Boolean)
class Boolean:
    @op(name="boolean-equal")
    def bool_equals(lhs: typing.Optional[bool], rhs: typing.Optional[bool]) -> bool:  # type: ignore
        return lhs == rhs

    @op(name="boolean-notEqual")
    def bool_notEquals(lhs: typing.Optional[bool], rhs: typing.Optional[bool]) -> bool:  # type: ignore
        return lhs != rhs

    @op(name="and")
    def bool_and(lhs: bool, rhs: typing.Optional[bool]) -> bool:  # type: ignore
        return lhs and bool(rhs)

    @op(name="or")
    def bool_or(lhs: bool, rhs: typing.Optional[bool]) -> bool:  # type: ignore
        return lhs or bool(rhs)

    @op(name="boolean-not")
    def bool_not(bool: bool) -> bool:  # type: ignore
        return not bool


types.Boolean.instance_class = Boolean


def none_coalesce_output_type(input_types):
    if types.NoneType().assign_type(input_types["lhs"]):
        return input_types["rhs"]
    return types.union(input_types["lhs"], input_types["rhs"])


@op(name="none-coalesce", output_type=none_coalesce_output_type)
def none_coalesce(lhs: typing.Any, rhs: typing.Any):
    # TODO: This logic is really complicated in Weave0.
    return lhs or rhs



import numpy as np
import pandas as pd
import json
import itertools
import typing

from . import projection_utils

from ._dict_utils import tag_aware_dict_val_for_escaped_key
from .. import box
from .. import weave_types as types
from ..api import Node, op, weave_class, OpVarArgs
from .. import errors
from .. import execute_fast
from .. import storage
from ..language_features.tagging import (
    tag_store,
    tagged_value_type,
    tagged_value_type_helpers,
)


import functools


def getitem_output_type(input_types, list_type=types.List):
    input_type_values = list(input_types.values())
    self_type = input_type_values[0]
    index_type = input_type_values[1]
    if self_type.object_type == types.UnknownType():
        # This happens when we're indexing an empty list
        obj_res = types.NoneType()
    else:
        obj_res = self_type.object_type
    if is_list_like(index_type):
        return list_type(obj_res)
    return obj_res


def general_picker(obj, key):
    return [row.get(key) for row in obj]


@weave_class(weave_type=types.List)
class List:
    @op(name="count", input_type={"arr": types.List(types.Any())})
    def count(arr: list[typing.Any]) -> int:  # type: ignore
        return len(arr)

    def __setitem__(self, k, v, action=None):
        # TODO: copy the whole list to an actual list so we can mutate!
        #    (not good, need to implement on ArrowTableList)
        self_list = list(self)
        self_list.__setitem__(k, v)
        return self_list

    @op(
        setter=__setitem__,
        input_type={
            "arr": types.List(types.Any()),
        },
        output_type=getitem_output_type,
    )
    def __getitem__(
        arr,
        index: typing.Optional[typing.Union[int, typing.List[typing.Optional[int]]]],
    ):
        if index == None:
            return None
        if isinstance(index, int):
            try:
                return arr.__getitem__(index)
            except IndexError:
                return None
        # otherwise we have a list-like (list or AWL)
        indexes = typing.cast(typing.List[typing.Optional[int]], index)
        result: list = []
        for i in indexes:
            if i == None:
                result.append(None)
            else:
                try:
                    result.append(arr.__getitem__(i))
                except IndexError:
                    result.append(None)
        return result

    @op(
        name="filter",
        input_type={
            "arr": types.List(types.Any()),
            "filterFn": lambda input_types: types.Function(
                {"row": input_types["arr"].object_type, "index": types.Int()},
                types.optional(types.Boolean()),
            ),
        },
        output_type=lambda input_types: input_types["arr"],
    )
    def filter(arr, filterFn):
        call_results = execute_fast.fast_map_fn(arr, filterFn)
        result = []
        for row, keep in zip(arr, call_results):
            if keep:
                result.append(row)

        return result

    @op(
        name="sort",
        input_type={
            "arr": types.List(types.Any()),
            "compFn": lambda input_types: types.Function(
                {"row": input_types["arr"].object_type}, types.List(types.Any())
            ),
            "columnDirs": types.List(types.String()),
        },
        output_type=lambda input_types: input_types["arr"],
    )
    def sort(arr, compFn, columnDirs):
        def cmp(a, b):
            a_comp_res = a[0]
            b_comp_res = b[0]
            for a_res, b_res, c_dir in zip(a_comp_res, b_comp_res, columnDirs):
                dir_adjust = -1 if c_dir == "desc" else 1
                if a_res == None and b_res == None:
                    continue
                elif a_res == None:
                    return -1 * dir_adjust
                elif b_res == None:
                    return 1 * dir_adjust

                try:
                    lt = a_res < b_res
                except TypeError:
                    a_res = str(a_res)
                    b_res = str(b_res)
                    lt = a_res < b_res

                if lt:
                    return -1 * dir_adjust
                elif a_res > b_res:
                    return 1 * dir_adjust
            return 0

        call_results = execute_fast.fast_map_fn(arr, compFn)
        sortable_results = zip(call_results, arr)
        sorted_results = sorted(sortable_results, key=functools.cmp_to_key(cmp))
        return [res[1] for res in sorted_results]

    @op(
        name="map",
        input_type={
            "arr": types.List(types.Any()),
            "mapFn": lambda input_types: types.Function(
                {"row": input_types["arr"].object_type, "index": types.Int()},
                types.Any(),
            ),
        },
        output_type=lambda input_types: types.List(input_types["mapFn"].output_type),
    )
    def map(arr, mapFn):
        return execute_fast.fast_map_fn(arr, mapFn)

    # TODO: This should be hidden!
    @op(
        name="_listmap",
        input_type={
            "arr": types.List(types.Any()),
            "mapFn": lambda input_types: types.Function(
                {"row": input_types["arr"].object_type}, types.Any()
            ),
        },
        output_type=lambda input_types: types.List(input_types["mapFn"].output_type),
    )
    def _listmap(arr, mapFn):
        return execute_fast.fast_map_fn(arr, mapFn)

    @op(
        name="groupby",
        input_type={
            "arr": types.List(types.Any()),
            "groupByFn": lambda input_types: types.Function(
                {"row": input_types["arr"].object_type}, types.Any()
            ),
        },
        output_type=lambda input_types: types.List(
            tagged_value_type.TaggedValueType(
                types.TypedDict(
                    {
                        "groupKey": input_types["groupByFn"].output_type,
                    }
                ),
                input_types["arr"],
            )
        ),
    )
    def groupby(arr, groupByFn):
        call_results = execute_fast.fast_map_fn(arr, groupByFn)
        result = {}
        # Start a new tagging context since we don't want the tags messing
        # with the groupby keys
        with tag_store.new_tagging_context():
            for row, group_key_items in zip(arr, call_results):
                group_key_s = json.dumps(storage.to_python(group_key_items))
                # logging.error("groupby group_key_s: %s", group_key_s)
                if group_key_s not in result:
                    result[group_key_s] = (group_key_items, [])
                result[group_key_s][1].append(row)
        grs = []
        for group_result in result.values():
            item = box.box(group_result[1])
            tag_store.add_tags(item, {"groupKey": group_result[0]})
            grs.append(item)
        return grs

    @op(
        name="offset",
        input_type={"arr": types.List(types.Any()), "offset": types.Number()},
        output_type=lambda input_types: input_types["arr"],
    )
    def offset(arr, offset):
        return arr[offset:]

    @op(
        name="limit",
        input_type={"arr": types.List(types.Any()), "limit": types.Number()},
        output_type=lambda input_types: input_types["arr"],
    )
    def limit(arr, limit):
        return arr[:limit]

    @op(
        name="dropna",
        input_type={"arr": types.List(types.Any())},
        output_type=lambda input_types: types.List(
            types.non_none(input_types["arr"].object_type)
            if types.non_none(input_types["arr"].object_type) != types.Invalid()
            else types.NoneType()
        ),
    )
    def dropna(arr):
        return [i for i in arr if i != None]

    @op(
        name="concat",
        input_type={
            "arr": types.List(types.union(types.NoneType(), types.List(types.Any())))
        },
        output_type=lambda input_types: input_types["arr"].object_type,
    )
    def concat(arr):
        res = []
        arr = [item for item in arr if item != None]
        for sublist in arr:
            if not tag_store.is_tagged(sublist):
                res.extend(sublist)
            else:
                tags = tag_store.get_tags(sublist)
                for i in sublist:
                    obj = box.box(i)
                    tag_store.add_tags(obj, tags)
                    res.append(obj)
        return res


def group_result_index_output_type(input_types):
    # THIS IS NO GOOD
    # TODO: need to fix Const type so we don't need this.
    self_type = input_types["self"]
    if isinstance(self_type, types.Const):
        return self_type.val_type.list.object_type
    else:
        return self_type.list.object_type


### TODO Move these ops onto List class and make part of the List interface


def _map_each_function_type(input_types: dict[str, types.Type]) -> types.Function:
    if types.List().assign_type(input_types["arr"]):
        return _map_each_function_type(
            {"arr": typing.cast(types.List, input_types["arr"]).object_type}
        )
    return types.Function({"row": input_types["arr"]}, types.Any())


def _map_each_output_type(input_types: dict[str, types.Type]):
    # propagate tags in output type
    map_each_function_type = input_types["mapFn"]
    output_type = typing.cast(types.Function, map_each_function_type).output_type

    def get_next_type(t: types.Type) -> types.Type:
        if types.List().assign_type(t):
            t_as_list = typing.cast(types.List, t)
            if tagged_value_type_helpers.is_tagged_value_type(t):
                t_as_tagged = typing.cast(tagged_value_type.TaggedValueType, t)
                return tagged_value_type.TaggedValueType(
                    t_as_tagged.tag,
                    types.List(
                        get_next_type(
                            typing.cast(types.List, t_as_tagged.value).object_type
                        )
                    ),
                )
            return types.List(get_next_type(t_as_list.object_type))
        elif tagged_value_type_helpers.is_tagged_value_type(t):
            t_as_tagged = typing.cast(tagged_value_type.TaggedValueType, t)
            return tagged_value_type.TaggedValueType(t_as_tagged.tag, output_type)
        return output_type

    return get_next_type(input_types["arr"])


def _map_each(arr: list, fn):
    existing_tags = [
        tag_store.get_tags(item) if tag_store.is_tagged(item) else None for item in arr
    ]

    if isinstance(arr, list) and len(arr) > 0 and isinstance(arr[0], list):
        result = [_map_each(item, fn) for item in arr]
    else:
        result = execute_fast.fast_map_fn(arr, fn)

    for i, item in enumerate(result):
        maybe_existing_tags = existing_tags[i]
        if maybe_existing_tags is not None:
            result[i] = tag_store.add_tags(box.box(item), maybe_existing_tags)

    return result


@op(
    name="mapEach",
    input_type={
        "arr": types.Any(),
        "mapFn": _map_each_function_type,
    },
    output_type=_map_each_output_type,
)
def map_each(arr, mapFn):
    arr_is_list = isinstance(arr, list)
    if not arr_is_list:
        arr = [arr]
    res = _map_each(arr, mapFn)
    if not arr_is_list:
        res = res[0]
    return res


def index_checkpoint_setter(arr, new_arr):
    return new_arr


@op(
    name="list-createIndexCheckpointTag",
    setter=index_checkpoint_setter,
    input_type={"arr": types.List(types.Any())},
    output_type=lambda input_types: types.List(
        tagged_value_type.TaggedValueType(
            types.TypedDict({"indexCheckpoint": types.Number()}),
            input_types["arr"].object_type,
        )
    ),
)
def list_indexCheckpoint(arr):
    # TODO: I think this will be inefficient for larger lists
    # or other data structures. Need to improve this.
    # if not isinstance(arr, list):
    #     return arr
    res = []
    for item in arr:
        item = box.box(item)
        tag_store.add_tags(item, {"indexCheckpoint": len(res)})
        res.append(item)
    return res


def is_list_like(list_type_or_node):
    if isinstance(list_type_or_node, Node):
        return types.is_list_like(list_type_or_node.type)
    return types.is_list_like(list_type_or_node)


def flatten_type(list_type):
    obj_type = list_type.object_type
    if is_list_like(obj_type):
        return flatten_type(obj_type)
    return types.List(obj_type)


def flatten_return_type(input_types):
    return flatten_type(input_types["arr"])


def _flatten(l):
    from ..ops_arrow import ArrowWeaveList
    from ..ops_arrow.arrow_tags import pushdown_list_tags

    if isinstance(l, list):
        tags = None
        if tag_store.is_tagged(l):
            tags = tag_store.get_tags(l)

        def tag_wrapper(item):
            if tags is not None:
                return tag_store.add_tags(box.box(item), tags)
            return item

        return sum((_flatten(tag_wrapper(o)) for o in l), [])
    elif isinstance(l, ArrowWeaveList):
        return pushdown_list_tags(l).to_pylist_tagged()
    else:
        return [l]


@op(
    name="flatten",
    input_type={"arr": types.List(types.Any())},
    output_type=flatten_return_type,
)
def flatten(arr):
    return _flatten(arr)


def unnest_return_type(input_types):
    arr_type = input_types["arr"]
    # if isinstance(arr_type, types.UnionType):
    #     return types.union(
    #         *(unnest_return_type({"arr": m}) for m in input_types["self"].members)
    #     )
    unnested_object_property_types = {}
    for k, v_type in arr_type.object_type.property_types.items():
        if types.List().assign_type(v_type):
            unnested_object_property_types[k] = v_type.object_type
        else:
            unnested_object_property_types[k] = v_type
    return types.List(types.TypedDict(unnested_object_property_types))


@op(
    name="unnest",
    input_type={"arr": types.List(types.TypedDict({}))},
    output_type=unnest_return_type,
)
def unnest(arr):
    if not arr:
        return arr
    list_cols = []
    # Very expensive to recompute type here. We already have it!
    # TODO: need a way to get argument types inside of resolver bodies.
    arr_type = types.TypeRegistry.type_of(arr)
    for k, v_type in arr_type.object_type.property_types.items():
        if types.List().assign_type(v_type):
            list_cols.append(k)
    if not list_cols:
        return arr
    return pd.DataFrame(arr).explode(list_cols).to_dict("records")


@op(
    name="unique",
    input_type={"arr": types.List(types.Any())},
    output_type=lambda input_types: input_types["arr"],
)
def unique(arr):
    if not arr:
        return arr
    res = np.unique(arr)
    return res.tolist()


def index_output_type(input_types):
    # THIS IS NO GOOD
    # TODO: need to fix Const type so we don't need this.
    self_type = input_types["arr"]
    if isinstance(self_type, types.Const):
        return self_type.val_type.object_type
    else:
        return self_type.object_type


def pick_output_type(input_types):
    # This is heinous. It handles mapped pick as well as regular pick.
    # Doesn't support all the fancy stuff with '*' that the frontend does yet.
    # TODO: fix, probably by a better mapped op implementation, and possible some
    # clearner way of implementing nice type logic.
    if not isinstance(input_types["key"], types.Const):
        return types.UnknownType()
    key = input_types["key"].val
    self_type = input_types["obj"]
    object_type = self_type
    is_list = False
    # Ew, this is the best way to determine if a Type looks like a list
    # (like pandas, sql list types).
    # TODO: fix
    if hasattr(self_type, "object_type"):
        object_type = self_type.object_type
        is_list = True
    if isinstance(object_type, types.Dict):
        output_type = object_type.object_type
    elif isinstance(object_type, types.TypedDict):
        property_types = object_type.property_types
        output_type = property_types.get(key)
        if output_type is None:
            output_type = types.none_type
    else:
        raise errors.WeaveInternalError(
            "pick received invalid input types: %s" % input_types
        )
    if is_list:
        output_type = types.List(output_type)
    return output_type


def list_return_type(input_types):
    return types.List(types.union(*input_types.values()))


@op(
    name="list",
    input_type=OpVarArgs(types.Any()),
    output_type=list_return_type,
)
def make_list(**l):
    return list(l.values())


def _join_output_type(input_types):
    arr1_prop_types = input_types["arr1"].object_type.property_types
    arr2_prop_types = input_types["arr2"].object_type.property_types
    prop_types = {}
    for k in set(arr1_prop_types.keys()).union(arr2_prop_types.keys()):
        prop_types[k] = types.List(types.union(arr1_prop_types[k], arr2_prop_types[k]))
    return types.List(types.TypedDict(prop_types))


@op(
    input_type={
        "arr1": types.List(types.TypedDict({})),
        "arr2": types.List(types.TypedDict({})),
        "keyFn1": lambda input_types: types.Function(
            {"row": input_types["arr1"].object_type}, types.Any()
        ),
        "keyFn2": lambda input_types: types.Function(
            {"row": input_types["arr2"].object_type}, types.Any()
        ),
    },
    output_type=_join_output_type,
)
def join2(arr1, arr2, keyFn1, keyFn2):  # type: ignore
    arr1_keys = execute_fast.fast_map_fn(arr1, keyFn1)
    arr2_keys = execute_fast.fast_map_fn(arr2, keyFn2)
    all_keys = set(arr1_keys).union(arr2_keys)
    arr1_lookup = dict(zip(arr1_keys, arr1))
    arr2_lookup = dict(zip(arr2_keys, arr2))
    results = []
    for k in all_keys:
        arr1_row = arr1_lookup[k]
        arr2_row = arr2_lookup[k]
        row_keys = set(arr1_row.keys()).union(arr2_row.keys())
        row = {}
        for rk in row_keys:
            row[rk] = [arr1_row.get(rk), arr2_row.get(rk)]
        results.append(row)
    return results


def _join_2_output_row_type(input_types):
    both_aliases_are_consts = isinstance(
        input_types["alias1"], types.Const
    ) and isinstance(input_types["alias2"], types.Const)
    arr1_obj_type = input_types["arr1"].object_type
    arr2_obj_type = input_types["arr2"].object_type
    if both_aliases_are_consts:
        alias_1 = input_types["alias1"].val
        alias_2 = input_types["alias2"].val
        result_row_type = types.TypedDict(
            {
                # Technically this optionality should depend on the `outers` flags
                # but making the implementation more loose for now to match Weave1
                alias_1: types.optional(arr1_obj_type),
                alias_2: types.optional(arr2_obj_type),
            }
        )
    else:
        result_row_type = types.Dict(
            types.String(),
            types.union(types.optional(arr1_obj_type), types.optional(arr2_obj_type)),
        )

    join_obj_type = types.optional(
        types.union(
            input_types["join1Fn"].output_type,
            input_types["join2Fn"].output_type,
        )
    )
    row_tag_type = types.TypedDict({"joinObj": join_obj_type})

    return tagged_value_type.TaggedValueType(row_tag_type, result_row_type)


def _join_2_output_type(input_types):
    return types.List(_join_2_output_row_type(input_types))


@op(
    name="join",
    input_type={
        "arr1": types.List(),
        "arr2": types.List(),
        "join1Fn": lambda input_types: types.Function(
            {"row": input_types["arr1"].object_type}, types.Any()
        ),
        "join2Fn": lambda input_types: types.Function(
            {"row": input_types["arr2"].object_type}, types.Any()
        ),
        "alias1": types.String(),
        "alias2": types.String(),
        "leftOuter": types.Boolean(),
        "rightOuter": types.Boolean(),
    },
    output_type=_join_2_output_type,
)
def join_2(arr1, arr2, join1Fn, join2Fn, alias1, alias2, leftOuter, rightOuter):
    table1_join_keys = execute_fast.fast_map_fn(arr1, join1Fn)
    table2_join_keys = execute_fast.fast_map_fn(arr2, join2Fn)

    jk_to_idx: dict[str, typing.Tuple[list[int], list[int], typing.Any]] = {}
    for i, jk in enumerate(table1_join_keys):
        if jk != None:
            safe_jk = json.dumps(box.unbox(jk))
            jk_to_idx.setdefault(safe_jk, ([], [], jk))[0].append(i)
    for i, jk in enumerate(table2_join_keys):
        if jk != None:
            safe_jk = json.dumps(box.unbox(jk))
            jk_to_idx.setdefault(safe_jk, ([], [], jk))[1].append(i)

    rows = []
    for idx1, idx2, jk in jk_to_idx.values():
        tag = {"joinObj": jk}
        if len(idx1) == 0 and rightOuter:
            idx1 = [None]
        if len(idx2) == 0 and leftOuter:
            idx2 = [None]
        for i1 in idx1:
            for i2 in idx2:
                row = {}
                if i1 != None:
                    row[alias1] = arr1[i1]
                else:
                    row[alias1] = None
                if i2 != None:
                    row[alias2] = arr2[i2]
                else:
                    row[alias2] = None
                rows.append(tag_store.add_tags(box.box(row), tag))
    return rows


def _join_all_output_type(input_types):
    arr_prop_types = input_types["arrs"].object_type.object_type.property_types
    prop_types = {}
    for k in arr_prop_types.keys():
        prop_types[k] = types.List(arr_prop_types[k])
    inner_type = types.TypedDict(prop_types)
    tag_type = types.TypedDict({"joinObj": input_types["joinFn"].output_type})
    tagged_type = tagged_value_type.TaggedValueType(tag_type, inner_type)
    return types.List(tagged_type)


def _all_element_keys(arrs: list[list[dict]]) -> set[str]:
    all_element_keys: set[str] = set([])
    for arr in arrs:
        for element in arr:
            all_element_keys = all_element_keys.union(element.keys())
    return all_element_keys


def _filter_none(arrs: list[typing.Any]) -> list[typing.Any]:
    return [a for a in arrs if a != None]


def _all_join_keys_mapping(
    arrs: list[list[dict]], join_keys: list[list[typing.Any]]
) -> typing.Tuple[set[typing.Any], list[dict[typing.Any, list[typing.Any]]]]:
    all_join_keys: set[typing.Any] = set([])
    join_key_mapping = []
    for arr, keys in zip(arrs, join_keys):
        arr_map: dict = {}
        for k, v in zip(keys, arr):
            if k != None:
                all_join_keys.add(k)
                if k in arr_map:
                    arr_map[k].append(v)
                else:
                    arr_map[k] = [v]
        join_key_mapping.append(arr_map)
    return all_join_keys, join_key_mapping


@op(
    name="joinAll",
    input_type={
        "arrs": types.List(types.optional(types.List(types.TypedDict({})))),
        "joinFn": lambda input_types: types.Function(
            {"row": input_types["arrs"].object_type.object_type}, types.Any()
        ),
    },
    output_type=_join_all_output_type,
)
def join_all(arrs, joinFn, outer: bool):
    # This is a pretty complicated op. See list.ts for the original implementation

    # First, filter out Nones
    arrs = _filter_none(arrs)

    # If nothing remains, simply return.
    if len(arrs) == 0:
        return []

    # Execute the joinFn on each of the arrays
    arrs_keys = [execute_fast.fast_map_fn(arr, joinFn) for arr in arrs]

    # Get the union of all the keys of all the elements
    all_element_keys = _all_element_keys(arrs)
    all_join_keys, join_key_mapping = _all_join_keys_mapping(arrs, arrs_keys)

    # Custom function for create a new row with the new value added
    def _add_val_to_joined_row(joined_row: dict, val: typing.Optional[typing.Any]):
        val = val or {}
        # This is a costly copy...
        copied = {k: [i for i in v] for k, v in joined_row.items()}
        for key in all_element_keys:
            copied[key].append(val.get(key, None))
        return copied

    default_vals = [None] if outer else []
    results = []

    # Each join key may result in several rows
    for k in all_join_keys:
        rows_for_join_key: list[dict] = [{k: [] for k in all_element_keys}]

        # Look through each arr mapping to find the rows for this join key
        for arr_lookup in join_key_mapping:
            # Get the values matching this key for this array
            arr_vals = arr_lookup.get(k, default_vals)

            # For each value, we will append it to every row in the existing list
            new_k_rows = []
            for v in arr_vals:
                for row in rows_for_join_key:
                    new_k_rows.append(_add_val_to_joined_row(row, v))

            # Update the main list
            rows_for_join_key = new_k_rows

        # Tag the elements and add them to the results
        for row in rows_for_join_key:
            row = tag_store.add_tags(box.box(row), {"joinObj": k})
            results.append(row)
    return results


@op(
    name="table-2DProjection",
    input_type={
        "table": types.List(types.optional(types.TypedDict({}))),
        "projectionAlgorithm": types.String(),
        "inputCardinality": types.String(),
        "inputColumnNames": types.List(types.String()),
        "algorithmOptions": types.TypedDict({}),
    },
    output_type=lambda input_types: types.List(
        types.TypedDict(
            {
                "projection": types.TypedDict(
                    {"x": types.Number(), "y": types.Number()}
                ),
                "source": input_types["table"].object_type,
            }
        )
    ),
)
def list_2Dprojection(
    table,
    projectionAlgorithm,
    inputCardinality,
    inputColumnNames,
    algorithmOptions,
):
    inputColumnNames = list(set(inputColumnNames))
    if len(inputColumnNames) == 0 or len(table) < 2:
        projection = [[0, 0] for row in table]
    else:
        embeddings = []
        for row in table:
            if inputCardinality == "single":
                embeddings.append(
                    tag_aware_dict_val_for_escaped_key(row, inputColumnNames[0])
                )
            else:
                embeddings.append(
                    [
                        tag_aware_dict_val_for_escaped_key(row, c)
                        for c in inputColumnNames
                    ]
                )
        np_array_of_embeddings = np.array(embeddings)
        np_projection = projection_utils.perform_2D_projection_with_timeout(
            np_array_of_embeddings,
            projectionAlgorithm,
            algorithmOptions,
        )
        projection = np_projection.tolist()
    return [
        {
            "projection": {
                "x": p[0],
                "y": p[1],
            },
            "source": row,
        }
        for p, row in zip(projection, table)
    ]


@op(
    name="table-projection2D",
    input_type={
        "table": types.List(types.optional(types.TypedDict({}))),
        "projectionAlgorithm": types.String(),
        "inputCardinality": types.String(),
        "inputColumnNames": types.List(types.String()),
        "algorithmOptions": types.TypedDict({}),
    },
    output_type=lambda input_types: types.List(
        types.TypedDict(
            {
                "projection": types.TypedDict(
                    {"x": types.Number(), "y": types.Number()}
                ),
                "source": input_types["table"].object_type,
            }
        )
    ),
)
def list_projection2D(
    table,
    projectionAlgorithm,
    inputCardinality,
    inputColumnNames,
    algorithmOptions,
):
    inputColumnNames = list(set(inputColumnNames))
    if len(inputColumnNames) == 0 or len(table) < 2:
        projection = [[0, 0] for row in table]
    else:
        embeddings = []
        for row in table:
            if inputCardinality == "single":
                embeddings.append(
                    tag_aware_dict_val_for_escaped_key(row, inputColumnNames[0])
                )
            else:
                embeddings.append(
                    [
                        tag_aware_dict_val_for_escaped_key(row, c)
                        for c in inputColumnNames
                    ]
                )
        np_array_of_embeddings = np.array(embeddings)
        np_projection = projection_utils.perform_2D_projection_with_timeout(
            np_array_of_embeddings, projectionAlgorithm, algorithmOptions
        )
        projection = np_projection.tolist()
    return [
        {
            "projection": {
                "x": p[0],
                "y": p[1],
            },
            "source": row,
        }
        for p, row in zip(projection, table)
    ]


@op(
    name="joinToStr",
    input_type={"arr": types.List(types.Any()), "sep": types.String()},
    output_type=types.String(),
)
def join_to_str(arr, sep):
    # This logic matches Weave0 (nulls are treated as empty strings)
    return sep.join([str(i) if i is not None else "" for i in arr])


@op(
    name="contains",
    input_type={"arr": types.List(types.Any()), "element": types.Any()},
    output_type=types.Boolean(),
)
def contains(arr, element):
    return element in arr


@op(
    pure=False,
    output_type=lambda input_types: input_types["arr"],
    name="list-randomlyDownsample",
)
def sample(arr: list[typing.Any], n: int):
    if n < 0:
        raise ValueError("n must be non-negative")
    elif n >= len(arr):
        return arr

    zeros = np.zeros(len(arr), dtype=bool)
    indices = np.random.choice(len(zeros), n, replace=False)
    zeros[indices] = True

    return [arr[i] for i in range(len(arr)) if zeros[i]]


ID_TYPE = types.UnionType(types.String(), types.Int())


def _typeddict_lookup_setter(arr, id, v):
    for i, row in enumerate(arr):
        if row.get("id") == id:
            break
    else:
        return arr
    arr[i] = v
    return arr


@op(
    name="listtypedict-lookup",
    setter=_typeddict_lookup_setter,
    input_type={
        "arr": types.List(types.TypedDict({"id": ID_TYPE})),
        "id": ID_TYPE,
    },
    output_type=getitem_output_type,
)
def typedict_lookup(arr, id):
    for row in arr:
        if row.get("id") == id:
            return row
    return None


def _object_lookup_setter(arr, id, v):
    for i, row in enumerate(arr):
        if row.id == id:
            break
    else:
        return arr
    arr[i] = v
    return arr


@op(
    name="listobject-lookup",
    setter=_object_lookup_setter,
    input_type={
        "arr": types.List(types.ObjectType(id=ID_TYPE)),
        "id": ID_TYPE,
    },
    output_type=getitem_output_type,
)
def object_lookup(arr, id):
    for row in arr:
        if row.id == id:
            return row
    return None


def _cross_product_output_type(input_types):
    obj_type = input_types["obj"]
    if isinstance(obj_type, types.TypedDict):
        new_prop_types = {}
        for k, v in obj_type.property_types.items():
            new_prop_types[k] = v.object_type
        return types.List(
            types.TypedDict(
                {
                    **new_prop_types,
                }
            )
        )
    return types.List(types.Dict(types.String(), types.String()))


@op(render_info={"type": "function"}, output_type=_cross_product_output_type)
def cross_product(obj: dict[str, list]):
    keys = list(obj.keys())
    values = list(obj.values())
    res = []
    for v in itertools.product(*values):
        res.append(dict(zip(keys, v)))
    return res



import typing
from ..api import op
from .. import artifact_fs


@op(name="FilesystemArtifactFile-directUrl")
def direct_url(file: artifact_fs.FilesystemArtifactFile) -> typing.Optional[str]:
    return file.artifact.direct_url(file.path)


@op(name="FilesystemArtifactFile-directUrlAsOf")
def direct_url_as_of(
    file: artifact_fs.FilesystemArtifactFile, asOf: int
) -> typing.Optional[str]:
    return file.artifact.direct_url(file.path)


@op(name="FilesystemArtifactFile-artifactVersion")
def file_artifact(
    file: artifact_fs.FilesystemArtifactFile,
) -> artifact_fs.FilesystemArtifact:
    return file.artifact



import dataclasses
import typing
import json
import time

import weave
from weave import artifact_base
from weave.artifact_fs import FilesystemArtifact
from weave.weave_types import Type

from langchain.vectorstores import VectorStore, FAISS
from langchain.vectorstores.base import VectorStoreRetriever
from langchain.chains.combine_documents.base import BaseCombineDocumentsChain
from langchain.chains.combine_documents.stuff import StuffDocumentsChain
from langchain.docstore.document import Document
from langchain.embeddings.base import Embeddings
from langchain.embeddings.openai import OpenAIEmbeddings
from langchain.llms import OpenAI
from langchain.chat_models.base import BaseChatModel
from langchain.chat_models import ChatOpenAI
from langchain.chat_models import ChatAnthropic
from langchain.chains import HypotheticalDocumentEmbedder
from langchain.chains.llm import LLMChain
from langchain.llms.base import BaseLLM
from langchain.llms.openai import BaseOpenAI
from langchain.callbacks.tracers.base import BaseTracer
from langchain.chains import RetrievalQA
from langchain.chains.base import Chain
from langchain.chains.retrieval_qa.base import BaseRetrievalQA
from langchain.base_language import BaseLanguageModel
from langchain.schema import BaseRetriever
from langchain.prompts.base import StringPromptTemplate
from langchain.prompts import PromptTemplate, BasePromptTemplate
from langchain.prompts import BaseChatPromptTemplate, ChatPromptTemplate
from langchain.prompts.chat import (
    BaseMessagePromptTemplate,
    BaseStringMessagePromptTemplate,
    ChatMessagePromptTemplate,
    HumanMessagePromptTemplate,
    AIMessagePromptTemplate,
    SystemMessagePromptTemplate,
)
from langchain.callbacks.tracers.schemas import Run

from ...ops_domain import trace_tree
from ... import storage
from . import util

import faiss


class WeaveTracer(BaseTracer):
    run: typing.Optional[Run]

    def _persist_run(self, run: "Run") -> None:
        self.run = run


@dataclasses.dataclass(frozen=True)
class DocumentType(weave.types.ObjectType):
    instance_classes = Document

    metadata: weave.types.Type = weave.types.TypedDict()

    def property_types(self) -> dict[str, Type]:
        return {"page_content": weave.types.String(), "metadata": self.metadata}


class VectorStoreType(weave.types.Type):
    instance_classes = VectorStore


# TODO: LangChain uses pydantic, we can infer object types everywhere...
class EmbeddingsType(weave.types.ObjectType):
    instance_classes = Embeddings


class OpenAIEmbeddingsType(EmbeddingsType):
    instance_classes = OpenAIEmbeddings


@dataclasses.dataclass(frozen=True)
class BaseRetrieverType(weave.types.ObjectType):
    instance_classes = BaseRetriever


@dataclasses.dataclass(frozen=True)
class VectorStoreRetrieverType(BaseRetrieverType):
    instance_classes = VectorStoreRetriever

    vectorstore: weave.types.Type = weave.types.Any()

    def property_types(self):
        return {
            "vectorstore": self.vectorstore,
            "search_type": weave.types.String(),
            # TODO: wrong, this is variable.
            "search_kwargs": weave.types.TypedDict(),
        }


@dataclasses.dataclass(frozen=True)
class BasePromptTemplateType(weave.types.ObjectType):
    instance_classes = BasePromptTemplate


@dataclasses.dataclass(frozen=True)
class StringPromptTemplateType(BasePromptTemplateType):
    instance_classes = StringPromptTemplate


@dataclasses.dataclass(frozen=True)
class PromptTemplateType(StringPromptTemplateType):
    instance_classes = PromptTemplate

    def property_types(self) -> dict[str, Type]:
        prop_types = super().property_types()
        prop_types.update(
            {
                "input_variables": weave.types.List(weave.types.String()),
                "template": weave.types.String(),
                "template_format": weave.types.String(),
                "validate_template": weave.types.Boolean(),
            }
        )
        return prop_types


@dataclasses.dataclass(frozen=True)
class BaseChatPromptTemplateType(BasePromptTemplateType):
    instance_classes = BaseChatPromptTemplate


@dataclasses.dataclass(frozen=True)
class ChatPromptTemplateType(BaseChatPromptTemplateType):
    instance_classes = ChatPromptTemplate

    messages: weave.types.Type = weave.types.Any()

    def property_types(self) -> dict[str, Type]:
        prop_types = super().property_types()
        prop_types.update(
            {
                "input_variables": weave.types.List(weave.types.String()),
                "messages": self.messages,
            }
        )
        return prop_types


@dataclasses.dataclass(frozen=True)
class BaseMessagePromptTemplateType(weave.types.ObjectType):
    instance_classes = BaseMessagePromptTemplate


@dataclasses.dataclass(frozen=True)
class BaseStringMessagePromptTemplateType(BaseMessagePromptTemplateType):
    instance_classes = BaseStringMessagePromptTemplate

    prompt: weave.types.Type = weave.types.Any()
    additional_kwargs: weave.types.Type = weave.types.Any()

    def property_types(self):
        return {
            "prompt": self.prompt,
            "additional_kwargs": self.additional_kwargs,
        }


@dataclasses.dataclass(frozen=True)
class ChatMessagePromptTemplateType(BaseStringMessagePromptTemplateType):
    instance_classes = ChatMessagePromptTemplate

    def property_types(self):
        prop_types = super().property_types()
        prop_types.update(
            {
                "role": weave.types.String(),
            }
        )
        return prop_types


@dataclasses.dataclass(frozen=True)
class HumanMessagePromptTemplateType(BaseStringMessagePromptTemplateType):
    instance_classes = HumanMessagePromptTemplate


@dataclasses.dataclass(frozen=True)
class AIMessagePromptTemplateType(BaseStringMessagePromptTemplateType):
    instance_classes = AIMessagePromptTemplate


@dataclasses.dataclass(frozen=True)
class SystemMessagePromptTemplateType(BaseStringMessagePromptTemplateType):
    instance_classes = SystemMessagePromptTemplate


@dataclasses.dataclass(frozen=True)
class ChainType(weave.types.ObjectType):
    instance_classes = Chain


# Hmm... this uses multiple inheritance in langchain, its both ChainType
# and EmbeddingsType. Weave doesn't support that, prefering type dispatch.
@dataclasses.dataclass(frozen=True)
class HyptheticalDocumentEmbedderType(EmbeddingsType):
    instance_classes = HypotheticalDocumentEmbedder

    base_embeddings: weave.types.Type = weave.types.Any()
    llm_chain: weave.types.Type = weave.types.Any()

    def property_types(self):
        return {
            "base_embeddings": self.base_embeddings,
            "llm_chain": self.llm_chain,
        }


@dataclasses.dataclass(frozen=True)
class LLMChainType(ChainType):
    instance_classes = LLMChain

    prompt: weave.types.Type = weave.types.Any()
    llm: weave.types.Type = weave.types.Any()

    def property_types(self):
        return {
            "prompt": self.prompt,
            "llm": self.llm,
            "output_key": weave.types.String(),
        }


@dataclasses.dataclass(frozen=True)
class BaseLanguageModelType(ChainType):
    instance_classes = BaseLanguageModel


@dataclasses.dataclass(frozen=True)
class BaseLLMType(BaseLanguageModelType):
    instance_classes = BaseLLM


@dataclasses.dataclass(frozen=True)
class BaseOpenAIType(BaseLLMType):
    instance_classes = BaseOpenAI

    def property_types(self) -> dict[str, Type]:
        return {
            "model_name": weave.types.String(),
            "temperature": weave.types.Float(),
        }


@dataclasses.dataclass(frozen=True)
class OpenAIType(BaseOpenAIType):
    instance_classes = OpenAI


@dataclasses.dataclass(frozen=True)
class BaseChatModelType(BaseLanguageModelType):
    instance_classes = BaseChatModel


@dataclasses.dataclass(frozen=True)
class ChatOpenAIType(BaseChatModelType):
    instance_classes = ChatOpenAI

    def property_types(self) -> dict[str, Type]:
        return {
            "model_name": weave.types.String(),
            "temperature": weave.types.Float(),
        }


@dataclasses.dataclass(frozen=True)
class ChatAnthropicType(BaseChatModelType):
    instance_classes = ChatAnthropic

    def property_types(self) -> dict[str, Type]:
        return {
            "model": weave.types.String(),
            "temperature": weave.types.Float(),
        }


@dataclasses.dataclass(frozen=True)
class BaseCombineDocumentsChainType(ChainType):
    instance_classes = BaseCombineDocumentsChain

    def property_types(self):
        return {
            "input_key": weave.types.String(),
            "output_key": weave.types.String(),
        }


@dataclasses.dataclass(frozen=True)
class StuffDocumentsChainType(BaseCombineDocumentsChainType):
    instance_classes = StuffDocumentsChain

    llm_chain: weave.types.Type = weave.types.Any()
    document_prompt: weave.types.Type = weave.types.Any()

    def property_types(self):
        prop_types = super().property_types()
        prop_types.update(
            {
                "llm_chain": self.llm_chain,
                "document_prompt": self.document_prompt,
                "document_variable_name": weave.types.String(),
                "document_separator": weave.types.String(),
            }
        )
        return prop_types


@dataclasses.dataclass(frozen=True)
class BaseRetrievalQAType(ChainType):
    instance_classes = BaseRetrievalQA

    combine_documents_chain: weave.types.Type = weave.types.Any()

    def property_types(self):
        return {
            "combine_documents_chain": self.combine_documents_chain,
            "input_key": weave.types.String(),
            "output_key": weave.types.String(),
            "return_source_documents": weave.types.Boolean(),
        }


@dataclasses.dataclass(frozen=True)
class RetrievalQAType(BaseRetrievalQAType):
    instance_classes = RetrievalQA

    retriever: weave.types.Type = weave.types.Any()

    def property_types(self):
        prop_types = super().property_types()
        prop_types.update(
            {
                "retriever": self.retriever,
            }
        )
        return prop_types


class FAISSType(VectorStoreType):
    instance_classes = FAISS

    def save_instance(
        self, obj: FAISS, artifact, name
    ) -> typing.Union[list[str], artifact_base.ArtifactRef, None]:
        # Langchain is inefficient, it resaves all the documents.
        # But in weave we've probably already saved them, so we'd prefer to just store a ref.
        # TODO: replace with weave implementation
        with artifact.new_dir(name) as dir:
            obj.save_local(dir)
        return None

    def load_instance(
        self,
        artifact: FilesystemArtifact,
        name: str,
        extra: typing.Union[list[str], None] = None,
    ) -> typing.Any:
        # TODO: LangChain requires Embeddings when loading, but it drops the embeddings
        # object!
        return FAISS.load_local(artifact.path(name), OpenAIEmbeddings())


@weave.op()
def faiss_from_documents(
    documents: list[Document],
    embeddings: Embeddings,
) -> FAISS:
    return FAISS.from_documents(documents, embeddings)


class DocumentEmbedding(typing.TypedDict):
    document: Document
    # TODO: np array
    embedding: list[float]
    # I made this a string because there's no way to tell Plot
    # its categorical at the moment.
    cluster: str


@weave.weave_class(weave_type=FAISSType)
class FaissOps:
    @weave.op()
    def document_embeddings(vector_store: FAISS) -> list[DocumentEmbedding]:
        embeddings = vector_store.index.reconstruct_n()
        kmeans = faiss.Kmeans(embeddings.shape[1], 20, niter=20, verbose=False)
        kmeans.train(embeddings)
        cluster_distances, cluster_ids = kmeans.index.search(embeddings, 1)
        cluster_ids = cluster_ids.tolist()
        result: list[DocumentEmbedding] = []
        for i, embedding in enumerate(embeddings):
            result.append(
                {
                    "document": vector_store.docstore.search(
                        vector_store.index_to_docstore_id[i]
                    ),
                    "embedding": embedding.tolist(),
                    "cluster": str(cluster_ids[i][0]),
                }
            )
        return result


@weave.op()
def similarity_search(vector_store: VectorStore, query: str) -> list[Document]:
    return vector_store.similarity_search(query)


@weave.op(render_info={"type": "function"})
def openai_embeddings() -> OpenAIEmbeddings:
    return OpenAIEmbeddings()


@weave.op(render_info={"type": "function"})
def openai(model_name: str, temperature: float) -> OpenAI:
    return OpenAI(model_name=model_name, temperature=temperature)


@weave.op(render_info={"type": "function"})
def chat_openai(model_name: str, temperature: float) -> ChatOpenAI:
    return ChatOpenAI(model_name=model_name, temperature=temperature)


@weave.op(render_info={"type": "function"})
def chat_anthropic(model_name: str, temperature: float) -> ChatAnthropic:
    return ChatAnthropic(model=model_name, temperature=temperature)


@weave.op(render_info={"type": "function"})
def llm_chain(llm: BaseLanguageModel, prompt: BasePromptTemplate) -> LLMChain:
    return LLMChain(llm=llm, prompt=prompt)


@weave.op(render_info={"type": "function"})
def hypothetical_document_embedder(
    base_embeddings: Embeddings, llm_chain: LLMChain
) -> HypotheticalDocumentEmbedder:
    return HypotheticalDocumentEmbedder(
        base_embeddings=base_embeddings, llm_chain=llm_chain
    )


@weave.op(render_info={"type": "function"})
def retrieval_qa_from_chain_type(
    language_model: BaseLanguageModel, chain_type: str, retriever: VectorStore
) -> RetrievalQA:
    retriever = retriever.as_retriever()
    return RetrievalQA.from_chain_type(
        llm=language_model,
        chain_type=chain_type,
        retriever=retriever,
    )


@weave.weave_class(weave_type=BaseChatModelType)
class BaseChatModelOps:
    @weave.op()
    def predict(self: BaseChatModel, text: str) -> str:
        if text == None:
            # TODO: weave engine doesn't handle nullability on args that aren't the first
            return None  # type: ignore
        return self.predict(text)


@weave.weave_class(weave_type=BaseLLMType)
class BaseLLMOps:
    @weave.op()
    def predict(self: BaseLLM, text: str) -> str:
        if text == None:
            # TODO: weave engine doesn't handle nullability on args that aren't the first
            return None  # type: ignore
        return self.predict(text)


ChainTypeVar = typing.TypeVar("ChainTypeVar", bound=Chain)


@weave.type()
class ChainRunResult(typing.Generic[ChainTypeVar]):
    chain: ChainTypeVar
    query: str
    result: str
    latency: float
    trace: trace_tree.WBTraceTree


@weave.weave_class(weave_type=ChainType)
class ChainOps:
    @weave.op(
        output_type=lambda input_type: ChainRunResult.WeaveType(  # type: ignore
            chain=input_type["chain"]
        )
    )
    def run(chain: Chain, query: str):
        if query == None:
            # TODO: weave engine doesn't handle nullability on args that aren't the first
            return None  # type: ignore
        tracer = WeaveTracer()
        start_time = time.time()
        result = chain.run(query, callbacks=[tracer])
        latency = time.time() - start_time
        lc_run = tracer.run
        if lc_run is None:
            raise ValueError("LangChain run was not recorded")
        span = util.safely_convert_lc_run_to_wb_span(lc_run)
        if span is None:
            raise ValueError("Could not convert LangChain run to Weave span")

        # Returns the langchain trace as part of the result... not really what we
        # want.
        return ChainRunResult(
            chain=chain,
            query=query,
            result=result,
            latency=latency,
            trace=trace_tree.WBTraceTree(json.dumps(storage.to_python(span)["_val"])),
        )


# @weave.weave_class(weave_type=BaseRetrievalQAType)
# class BaseRetrievalQAOps:
#     @weave.op()
#     def run(chain: BaseRetrievalQA, query: str) -> RunResult:
#         if query == None:
#             # TODO: weave engine doesn't handle nullability on args that aren't the first
#             return None  # type: ignore
#         tracer = WeaveTracer()
#         result = chain.run(query, callbacks=[tracer])
#         lc_run = tracer.run
#         if lc_run is None:
#             raise ValueError("LangChain run was not recorded")
#         span = util.safely_convert_lc_run_to_wb_span(lc_run)
#         if span is None:
#             raise ValueError("Could not convert LangChain run to Weave span")

#         # Returns the langchain trace as part of the result... not really what we
#         # want.
#         return RunResult(
#             result=result,
#             trace=trace_tree.WBTraceTree(json.dumps(storage.to_python(span)["_val"])),
#         )
